"""Command line interfaces."""

import argparse
import importlib
import os
import shutil
from typing import List, Optional

os.environ["ANSIBLE_COW_PATH"] = os.getenv(
    "ANSIBLE_COW_PATH", shutil.which("cowsay") or ""
)

from rich_argparse import ArgumentDefaultsRichHelpFormatter

from freva_deployment import __version__
from freva_deployment.ui.deployment_tui import tui
from freva_deployment.versions import VersionAction, display_versions

from ._compose import compose_parser
from ._config import config_parser
from ._deploy import BatchParser
from ._deploy import cli as deploy
from ._kubernets import kubernetes_parser
from ._migrate import cli as migrate
from ._migrate import create_parser as migrate_parser

__all__ = ["deploy", "migrate"]


def __getattr__(name):
    return getattr(importlib.import_module(f"._{name}", __name__), "cli")


def main_cli(argv: Optional[List[str]] = None) -> None:
    """Construct command line argument parser."""
    app = argparse.ArgumentParser(
        prog="deploy-freva",
        description="Run the freva deployment",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    app.set_defaults(cli=tui)
    app.add_argument(
        "-v", "--verbose", action="count", help="Verbosity level", default=0
    )
    app.add_argument(
        "-V",
        "--version",
        action=VersionAction,
        version="[b][red]%(prog)s[/red] {version}[/b]{services}".format(
            version=__version__, services=display_versions()
        ),
    )
    app.add_argument(
        "--cowsay",
        action="store_true",
        help="Let the cow speak!",
        default=False,
    )

    subparser = app.add_subparsers(
        required=False,
    )
    _ = BatchParser(
        subparser.add_parser(
            name="cmd",
            help="Run deployment in batch mode.",
            description="Run deployment in batch mode.",
            formatter_class=ArgumentDefaultsRichHelpFormatter,
        )
    )
    config_parser(
        parser=subparser.add_parser(
            name="config",
            help="Create and inspect freva configuration.",
            formatter_class=ArgumentDefaultsRichHelpFormatter,
        )
    )
    compose_parser(
        parser=subparser.add_parser(
            name="compose",
            help="Create a compose file.",
            formatter_class=ArgumentDefaultsRichHelpFormatter,
        )
    )
    kubernetes_parser(
        parser=subparser.add_parser(
            name="kubernetes",
            help="Create a k8s manifests for deployment.",
            formatter_class=ArgumentDefaultsRichHelpFormatter,
        )
    )
    migrate_parser(
        parser=subparser.add_parser(
            name="migrate",
            help="Utilities to handle migrations from the legacy freva.",
            formatter_class=ArgumentDefaultsRichHelpFormatter,
        )
    )

    args = app.parse_args(argv)
    args.cli(args)



#!/usr/bin/env python3
import argparse
import platform
import tarfile
import urllib.request as req
from pathlib import Path
from tempfile import NamedTemporaryFile


def reporthook(count: float, block_size: float, total_size: float) -> None:
    """Print the download status."""
    if count == 0:
        return
    frac = count * block_size / total_size
    percent = int(100 * frac)
    bars = "#" * int(frac * 40)
    msg = "Downloading: [{0:<{1}}] | {2}% Completed".format(
        bars, 40, round(percent, 2)
    )
    print(msg, end="\r", flush=True)
    if frac >= 1:
        print()


def _url_retrieve(conda_url: str, target: str) -> None:
    print("Retrieving conda install script: {}".format(conda_url))
    try:
        req.urlretrieve(
            str(conda_url),
            filename=target,
            reporthook=reporthook,
        )
    except Exception as error:
        print(error)


def _mamba_forge(dest_dir: Path) -> None:
    target = dest_dir / "conda.sh"
    _url_retrieve(
        "https://github.com/conda-forge/miniforge/releases/latest/"
        f"download/Miniforge3-{platform.system()}-{platform.machine()}.sh",
        str(target),
    )
    target.chmod(0o755)


def _micromamba(dest_dir: Path) -> None:
    system = platform.system().lower()
    plt = platform.machine()
    target = dest_dir / "bin" / "micromamba"
    if system != "linux":
        raise ValueError("Only Linux based deployment is supported.")
    if plt.startswith("x86"):
        plt = "64"
    with NamedTemporaryFile(suffix=".tar") as temp_f:
        _url_retrieve(
            f"https://micro.mamba.pm/api/micromamba/linux-{plt}/latest",
            temp_f.name,
        )
        with tarfile.open(temp_f.name, mode="r:bz2") as tar:
            member = tar.getmember("bin/micromamba")
            print("🔧 Extracting: bin/micromamba")
            tar.extract(member, path=str(dest_dir))
    target.chmod(0o755)


def download(mamba: str = "micromamba", dest_dir: Path = Path("/tmp")) -> None:
    """Download the conda forge install script."""
    dest_dir.mkdir(parents=True, exist_ok=True)
    if "forge" in mamba:
        _mamba_forge(dest_dir)
    else:
        _micromamba(dest_dir)


def _cli() -> argparse.ArgumentParser:

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "target_path", type=Path, help="Where to download the mamba exe."
    )
    parser.add_argument(
        "--type",
        "-t",
        type=str,
        default="micromamba",
        choices=("micromamba", "mambaforge"),
    )
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = _cli()
    download(args.type, args.target_path)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import io

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from contextlib import redirect_stderr, redirect_stdout
from datetime import date
from pathlib import Path

# from recommonmark.parser import CommonMarkParser

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath("../../src"))

from freva_deployment import __version__
from freva_deployment.cli import main_cli
from freva_deployment.utils import config_dir

toml_config_file = (
    Path(__file__).parent.parent
    / "assets"
    / "share"
    / "freva"
    / "deployment"
    / "config"
    / "inventory.toml"
)
toml_config = toml_config_file.read_text()


def get_cli_output(*args, format="md"):
    if args:
        cmd = list(args) + ["--help"]
    else:
        cmd = ["--help"]
    command = f"deploy-freva {' '.join(args)} --help"
    buf = io.StringIO()
    try:
        with redirect_stderr(buf), redirect_stdout(buf):
            main_cli(cmd)
    except SystemExit:
        pass
    output = buf.getvalue()
    if format == "md":
        return f"```console\n{command}\n{output}```"
    else:
        return f"{command}\n{output}\n"


cli_tui = get_cli_output()
cli_cmd = get_cli_output("cmd")
cli_mig = get_cli_output("migrate")
cli_config = get_cli_output("config")
# -- Project information -----------------------------------------------------

project = "freva-deployment"
copyright = f"{date.today().year}, DKRZ - CLINT"
author = "Climate Informatics and Technology"

# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "sphinx_togglebutton",
    "sphinxext.opengraph",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
html_logo = "_static/freva_owl.svg"
html_favicon = "_static/freva_owl.svg"
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_sidebars = {"pagename": []}
html_static_path = ["_static"]
html_context = {
    "github_user": "freva-org",
    "github_repo": "freva-admin",
    "github_version": "main",
    "doc_path": "docs",
}
html_sidebars = {
    "community/index": [
        "sidebar-nav-bs",
        "custom-template",
    ],  # This ensures we test for custom sidebars
}

html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/freva-org/freva-admin",
            "icon": "fa-brands fa-github",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/freva-deployment",
            "icon": "fa-custom fa-pypi",
        },
    ],
    "navbar_center": ["navbar-nav"],  # Add navigation links to the left sidebar
    "collapse_navigation": False,
    "navigation_depth": 4,
    "navbar_align": "left",
    "show_nav_level": 2,
    "secondary_sidebar_items": ["page-toc"],
}
html_sidebars = {"**": ["sidebar-nav-bs", "sidebar-ethical-ads"]}
html_meta = {
    "description": "Admin guide for freva.",
    "keywords": "freva, climate, data analysis, evaluation, framework, climate science",
    "author": "Freva Team",
    "og:title": "Freva – Free Evaluation System Framework",
    "og:description": "Admin guide for Freva.",
    "og:type": "website",
    "og:url": "https://freva-org.github.io/freva-admin/",
    "og:image": "https://freva-org.github.io/freva-admin/_images/freva_flowchart-new.png",
    "twitter:card": "summary_large_image",
    "twitter:title": "Freva – Evaluation System Framework",
    "twitter:description": "Search, analyse and evaluate climate model data.",
    "twitter:image": "https://freva-org.github.io/freva-admin/_images/freva_flowchart-new.png",
}

ogp_site_url = "https://freva-org.github.io/freva-admin"
opg_image = (
    "https://freva-org.github.io/freva-admin/_images/freva_flowchart-new.png",
)
ogp_type = "website"
ogp_custom_meta_tags = [
    '<meta name="twitter:card" content="summary_large_image">',
    '<meta name="keywords" content="freva, climate, data, evaluation, science, reproducibility">',
]
# -- MyST options ------------------------------------------------------------

# This allows us to use ::: to denote directives, useful for admonitions
myst_enable_extensions = ["colon_fence", "linkify", "substitution"]
myst_heading_anchors = 2
myst_substitutions = {
    "rtd": "[Read the Docs](https://readthedocs.org/)",
    "version": __version__,
    "cli_tui": cli_tui,
    "cli_cmd": cli_cmd,
    "cli_mig": cli_mig,
    "cli_compose": get_cli_output("compose"),
    "cli_k8s": get_cli_output("kubernetes"),
    "cli_config": cli_config,
    "cli_config_get": get_cli_output("config", "get"),
    "cli_config_set": get_cli_output("config", "get"),
    "toml_config": f"```toml\n{toml_config}\n```",
}
myst_url_schemes = {
    "http": None,
    "https": None,
    "release": f"https://github.com/freva-org/freva-admin/releases/tag/v{release}",
    "exe": f"https://github.com/freva-org/freva-admin/releases/download/v{release}/deploy-freva-v{release}-{{{{path}}}}",
}
# Substitutions
rst_prolog = """
.. version replace:: {version}
""".format(
    version=__version__,
)

cli_output_dir = Path(__file__).parent / "_generated"
cli_output_dir.mkdir(exist_ok=True, parents=True)

(cli_output_dir / "compose.txt").write_text(
    get_cli_output("compose", format="rst")
)

(cli_output_dir / "k8s.txt").write_text(
    get_cli_output("kubernetes", format="rst")
)


# ReadTheDocs has its own way of generating sitemaps, etc.
if not os.environ.get("READTHEDOCS"):
    extensions += ["sphinx_sitemap"]

    html_baseurl = os.environ.get("SITEMAP_URL_BASE", "http://127.0.0.1:8000/")
    sitemap_locales = [None]
    sitemap_url_scheme = "{link}"

# specifying the natural language populates some key tags
language = "en"



"""Command line interface for the configuration."""

from __future__ import annotations

import argparse
from pathlib import Path
from typing import Any, Optional

import tomlkit
from rich.console import Console
from rich_argparse import ArgumentDefaultsRichHelpFormatter

from freva_deployment import __version__

from ..utils import config_dir, load_config


def _get_config(parser: argparse.Namespace) -> None:
    """Display config."""
    error_console = Console(markup=True, force_terminal=True, stderr=True)
    std_console = Console(markup=True, force_terminal=True)
    try:
        cfg = load_config(parser.config_file)
    except Exception as error:
        if parser.verbose > 0:
            raise
        error_console.print(f"[b]:warning:  {error}[/b]")
        raise SystemExit(1)
    if not parser.keys:
        if parser.raw:
            print(parser.config_file.read_text())
        else:
            std_console.print(parser.config_file.read_text())
        return
    for key in parser.keys:
        section, _, value = key.partition(".")
        try:
            sec = cfg[section]
            if value:
                disp = sec[value]
            else:
                disp = sec
        except Exception as error:
            error_console.print(f"[b red]:warning:  {error}[/b red]")
            continue
    try:
        disp_dump = tomlkit.dumps(disp)
    except Exception:
        disp_dump = disp
    if parser.raw:
        print(disp_dump)
    else:
        std_console.print(disp_dump)


def _set_config(parser: argparse.Namespace) -> None:
    """Set config."""
    default_config_file = config_dir / "config" / "inventory.toml"

    error_console = Console(markup=True, force_terminal=True, stderr=True)
    if parser.config_file == default_config_file:
        error_console.print(
            "[b]:warning:  Using system default config, which will overridden "
            "on update.[/b]"
        )
    try:
        cfg = load_config(parser.config_file)
    except Exception as error:
        if parser.verbose > 0:
            raise
        error_console.print(f"[b]:warning:  {error}[/b]")
        raise SystemExit(1)
    default_config = tomlkit.loads(default_config_file.read_text())
    for keys, value in parser.values:
        section, _, key = keys.partition(".")
        if key:
            default = default_config.get(section, {}).get(key)
        else:
            default = default_config.get(section)
        try:
            cfg_value: Any = ""
            if isinstance(default, bool):
                if value.lower() == "false":
                    cfg_value = False
                else:
                    cfg_value = True
            elif isinstance(default, int) and value.isdigit():
                cfg_value = int(value)
            elif isinstance(default, float) and value.isdigit():
                cfg_value = float(value)
            elif isinstance(default, str):
                cfg_value = value
            else:
                cfg_value = tomlkit.loads(f"foo={value}")["foo"]
            if key:
                cfg[section][key] = cfg_value
            else:
                cfg[section] = cfg_value
        except Exception as error:
            if parser.verbose > 0:
                raise
            error_console.print(f"[b red]:warning:  {error}[/b red]")
            raise SystemExit(1)
    parser.config_file.write_text(tomlkit.dumps(cfg))


def config_parser(
    epilog: str = "", parser: Optional[argparse.ArgumentParser] = None
) -> argparse.ArgumentParser:
    """Construct command line argument parser."""
    parser = parser or argparse.ArgumentParser(
        prog="deploy-freva-config",
        description="Create and inspect freva configuration.",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
        epilog=epilog,
    )
    parser.add_argument(
        "-v", "--verbose", action="count", help="Verbosity level", default=0
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    subparsers = parser.add_subparsers(required=False)
    get_parser = subparsers.add_parser(
        "get",
        description="Inspect configuration values.",
        help=(
            "Use this command to query either the whole or parts of the "
            "deployment config."
        ),
        epilog=epilog,
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    get_parser.add_argument(
        "keys",
        nargs="*",
        help=(
            "Query keys. Can be the from of <section>.<key>, <key>, <section> "
            "if no keys are given the whole configuration is printed to stdout."
            " this can be useful for creating new config files"
        ),
    )
    get_parser.add_argument(
        "-c",
        "--config-file",
        type=Path,
        help="Path to ansible inventory file.",
        default=config_dir / "config" / "inventory.toml",
    )
    get_parser.add_argument(
        "-r", "--raw", help="Raw output", action="store_true", default=False
    )
    get_parser.add_argument(
        "-v", "--verbose", action="count", help="Verbosity level", default=0
    )

    set_parser = subparsers.add_parser(
        "set",
        description="Inspect configuration values.",
        help=(
            "Use this command to set/override values " "of the deployment config."
        ),
        epilog=epilog,
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    set_parser.add_argument(
        "-c",
        "--config-file",
        type=Path,
        help="Path to ansible inventory file.",
        default=config_dir / "config" / "inventory.toml",
    )
    set_parser.add_argument(
        "values",
        nargs=2,
        action="append",
        help=(
            "Key value pairs of the config. Can be in from of"
            "<section>.<key> value, <key> = value. Note: value needs "
            "to follow toml syntax"
        ),
    )
    set_parser.add_argument(
        "-v", "--verbose", action="count", help="Verbosity level", default=0
    )
    set_parser.set_defaults(cli=_set_config)
    get_parser.set_defaults(cli=_get_config)
    return parser



"""Call the freva tui."""

from __future__ import annotations

import argparse
import logging

from freva_deployment.deploy import DeployFactory
from freva_deployment.error import DeploymentError
from freva_deployment.logger import logger, set_log_level

from .main_window import MainApp


def tui(args: argparse.Namespace) -> None:
    """Create and run Text User Interface (TUI) for deployment."""
    set_log_level(args.verbose)
    try:
        main_app = MainApp()
        main_app.run()
    except (KeyboardInterrupt, Exception) as error:
        try:
            main_app.thread_stop.set()
            main_app.save_config_to_file(
                save_file=main_app._setup_form.inventory_file.value
            )
        except AttributeError:
            pass
        if isinstance(error, KeyboardInterrupt):
            raise
        msg = str(error)
        if "addwstr" in msg:
            msg = (
                "Cloud not display content, try " "increasing your terminal size"
            )
        if logger.getEffectiveLevel() < logging.INFO:
            logger.exception(error)
        else:
            logger.error("Exiting App: %s", msg)
        return
    setup = main_app.setup
    main_app.thread_stop.set()
    if setup:
        ask_pass = setup.pop("ask_pass")
        ssh_port = setup.pop("ssh_port")
        skip_version_check = setup.pop("skip_version_check", False)
        with DeployFactory(_cowsay=args.cowsay, **setup) as DF:
            try:
                DF.play(
                    ask_pass,
                    args.verbose,
                    ssh_port=ssh_port,
                    skip_version_check=skip_version_check,
                )
            except KeyboardInterrupt:
                raise SystemExit(130)
            except DeploymentError:
                raise SystemExit(1)



import locale
import os
import re
import sys

from rich import print as pprint


def main():
    """Run the main tui."""
    from freva_deployment.cli import main_cli

    path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "bin")
    for exe in ("cowsay", "cowsay.exe"):
        cow_exe = os.path.join(path, exe)
        if os.path.isfile(cow_exe):
            os.environ["ANSIBLE_COW_PATH"] = cow_exe
            break
    os.environ["PATH"] = os.environ["PATH"] + os.pathsep + path
    try:
        main_cli()
    except (KeyboardInterrupt, SystemExit):
        sys.exit(1)


if __name__ == "__main__":
    sys.argv[0] = re.sub(r"(-script\.pyw|\.exe)?$", "", sys.argv[0])
    try:
        locale.setlocale(locale.LC_ALL, "")
        dummy, encoding = locale.getlocale()
    except (locale.Error, ValueError):
        encoding = "unknown"
    if (encoding or "unknown").lower() not in ("utf-8", "utf8"):
        pprint(
            f"[red][b]WARNING[/b] - Your locale encoding is {encoding} "
            "Ansible might not work correctly. Consider changing your "
            "locale encoding (LC_ALL) to UTF-8[/red]"
        )
        locale.getlocale = lambda: ("UTF-8", "UTF-8")
        sys.getfilesystemencoding = lambda: "utf-8"
        os.environ["PYTHONIOENCODING"] = "utf-8"
        os.environ["LC_ALL"] = "C"
        os.environ["LANG"] = "C"
    main()



"""Command line interface for creating k8s manifest files."""

from __future__ import annotations

import argparse
import base64
import re
from base64 import b64encode
from pathlib import Path
from typing import Any, Dict, List, Optional

import namegenerator
import yaml
from rich_argparse import ArgumentDefaultsRichHelpFormatter

from freva_deployment import __version__

from ..deploy import DeployFactory
from ..logger import logger, set_log_level
from ..utils import RichConsole, asset_dir, config_dir

TASK = """---
- name: Render compose file locally only
  hosts: all
  connection: local
  gather_facts: yes

  tasks:

    - name: Render k8s manifests
      loop:
        "{{{{ templates }}}}"
      loop_control:
        loop_var: t
      template:
        src: "{asset_dir}/k8s-deployment/templates/{{{{ t }}}}"
        dest: "{{{{ out_dir }}}}/{{{{ t | basename | regex_replace('\\\\.j2$', '') }}}}"
"""


def get_ingress_hosts(
    project_name: str,
    services: List[str],
    domain: str,
    config: Dict[str, Any],
) -> List[Dict[str, str | int]]:
    """Define the ingress hosts."""
    lookup = {
        "freva-rest": [
            {
                "fqdn": f"freva-api.{domain}",
                "service": "freva-rest-server",
                "port": config["freva_rest"]["freva_rest_port"],
                "tls": False,
                "tls_secret": "web-cert-secret",
            },
            {
                "fqdn": f"freva-solr.{domain}",
                "service": "search-server",
                "port": 8983,
                "tls": False,
                "tls_secret": "web-cert-secret",
            },
        ],
        "web": [
            {
                "fqdn": f"freva-vault.{domain}",
                "service": "vault-server",
                "port": 5002,
                "tls": False,
                "tls_secret": "web-cert-secret",
            },
            {
                "fqdn": f"freva-db.{domain}",
                "service": "database-server",
                "port": 3306,
                "tls": False,
                "tls_secret": "web-cert-secret",
            },
        ],
    }
    fqdns: List[Dict[str, str | int]] = []
    for service in services:
        for cfg in lookup.get(service, []):
            if cfg not in fqdns:
                fqdns.append(cfg)
    return fqdns


def get_pvcs_from_services(
    services: List[str], config: Dict[str, Any]
) -> List[Dict[str, str]]:
    """Define the pvcs needed for each service."""
    lookup = {
        "freva-rest": [
            {
                "name": "solr-data",
                "mode": "ReadWriteOnce",
                "storage": config["solr"]["data_size"],
                "instance": "rest-api",
                "tier": "database",
                "component": "data",
            },
            {
                "name": "mongo-data",
                "mode": "ReadWriteOnce",
                "storage": config["mongo"]["data_size"],
                "instance": "mongo-server",
                "tier": "database",
                "component": "data",
            },
        ],
        "web": [
            {
                "name": "db-data",
                "mode": "ReadWriteOnce",
                "storage": config["mysql"]["data_size"],
                "instance": "database-server",
                "tier": "database",
                "component": "data",
            },
            {
                "name": "vault-data",
                "mode": "ReadWriteOnce",
                "storage": "500Mi",
                "instance": "web-app",
                "component": "data",
                "tier": "backend",
            },
            {
                "name": "web-data",
                "mode": "ReadWriteOnce",
                "storage": config["web"]["data_size"],
                "component": "data",
                "instance": "web-app",
                "tier": "backend",
            },
            {
                "name": "proxy-logs",
                "mode": "ReadWriteOnce",
                "storage": "500Mi",
                "component": "logs",
                "instance": "web-app",
                "tier": "frontend",
            },
        ],
    }
    lookup["web"] += lookup["freva-rest"]
    pvcs: List[Dict[str, str]] = []
    for service in services:
        for pvc_settings in lookup.get(service, []):
            if pvc_settings not in pvcs:
                pvcs.append(pvc_settings)
    return pvcs


def get_service_templates(services: List[str]) -> List[str]:
    """Assigns the templates needed for each the services."""
    lookup = {
        "freva-rest": ["12-solr.yaml.j2", "14-mongo.yaml.j2", "20-rest.yaml.j2"],
        "data-loader": ["11-redis.yaml.j2", "22-data-loader.yaml.j2"],
    }
    lookup["web"] = lookup["freva-rest"] + [
        "10-mysql.yaml.j2",
        "11-redis.yaml.j2",
        "13-vault.yaml.j2",
        "32-web-app.yaml.j2",
    ]
    templates: List[str] = [
        "00-namespace.yaml.j2",
        "01-secrets.yaml.j2",
        "02-pvcs.yaml.j2",
        "30-ingress.yaml.j2",
    ]
    for service in services:
        if service in lookup:
            templates += lookup[service]
    return sorted(set(templates))


def comment_entries(toml_str, entries_to_comment):
    lines = toml_str.splitlines()
    result = []
    for line in lines:
        stripped = line.lstrip()
        # If line starts with one of the target entries, comment it
        for entry in entries_to_comment:
            if re.match(rf'^\["{re.escape(entry)}",', stripped):
                indent = line[: len(line) - len(stripped)]
                result.append(f"{indent}#    {stripped}")
                break
        else:
            result.append(line)
    return "\n".join(result)


def create_manifest(args: argparse.Namespace) -> None:
    """Create the k8s manifests."""
    set_log_level(args.verbose)
    services = args.services
    with DeployFactory(
        steps=None,
        config_file=args.config_file,
        local_debug=False,
        gen_keys=True,
    ) as DF:
        out_dir = (
            (args.output or Path.cwd() / DF.project_name).expanduser().absolute()
        )
        out_dir.mkdir(exist_ok=True, parents=True)
        eval_conf_enc = b64encode(
            DF.create_eval_config().read_text().encode()
        ).decode()
        extra = {
            **{
                "eval_config_content": eval_conf_enc,
                "use_core": args.no_plugins is False,
                "redis_password": DF._create_random_passwd(30, 10),
                "redis_username": namegenerator.gen(),
                "ingress_enabled": True,
                "ingress_fqdns": [DF.cfg["web"]["project_website"]],
                "fqdn": DF.cfg["kubernetes"]["ingress"]["fqdn"],
                "out_dir": str(out_dir),
                "image_pull_policy": "IfNotPresent",
                "templates": get_service_templates(services),
                "pvcs": get_pvcs_from_services(
                    services, DF.cfg["kubernetes"]["pvc"]
                ),
                "resources": DF.cfg["kubernetes"]["resources"],
                "ingress_hosts": get_ingress_hosts(
                    DF.project_name,
                    services,
                    DF.cfg["kubernetes"]["ingress"]["fqdn"],
                    DF.cfg,
                ),
            },
            **DF.cfg["kubernetes"],
        }
        logger.info("Parsing configurations")
        logger.debug("Extra args for playbooks:\n%s", extra)
        inventory = yaml.safe_load(DF.parse_config(DF.steps, **extra))
        if args.no_plugins is True:
            web_config = base64.b64decode(
                inventory["web"]["vars"]["web_config_content"]
            ).decode()
            to_comment = ["Plugins", "History", "Result-Browser"]
            web_config = comment_entries(web_config, to_comment)
            inventory["web"]["vars"]["web_config_content"] = base64.b64encode(
                web_config.encode()
            ).decode()

        playbook = TASK.format(asset_dir=asset_dir)
        web_conf = (
            Path(inventory["core"]["vars"]["core_root_dir"])
            / "share"
            / "freva"
            / "web"
            / "freva_web.toml"
        )
        for key in inventory:
            inventory[key]["hosts"] = "localhost"
        inventory["web"]["vars"]["web_config_file"] = str(web_conf)
        logger.debug(yaml.safe_dump(inventory))
        DF._td.run_ansible_playbook(
            working_dir=asset_dir / "playbooks",
            playbook=playbook,
            inventory=inventory,
            verbosity=args.verbose,
        )

        RichConsole.rule("")
        RichConsole.print(
            (
                f"The k8s manifests have been created in [b]{out_dir}[/b].\n"
                f"You can apply them using kubctl.\n"
                "Before starting the web app make sure you have the core.\n"
                "library installed and prepared the web-directory structure."
                " on the HPC.\n"
                "You can do this by running the following command:\n\n"
                f"  [b]deploy-freva cmd -c {args.config_file} "
                "-t core pre-web --skip-version-check -g"
            )
        )
        RichConsole.rule("")


def kubernetes_parser(
    epilog: str = "", parser: Optional[argparse.ArgumentParser] = None
) -> None:
    """Construct command line argument parser."""
    parser = parser or argparse.ArgumentParser(
        prog="deploy-freva-kubernetes",
        description="Creat and inspect k8s manifests for freva deployment.",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
        epilog=epilog,
    )
    parser.add_argument(
        "-v", "--verbose", action="count", help="Verbosity level", default=0
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    parser.add_argument(
        "-c",
        "--config-file",
        type=Path,
        help="Path to ansible inventory file.",
        default=config_dir / "config" / "inventory.toml",
    )
    parser.add_argument(
        "-s",
        "--services",
        nargs="+",
        default=["db", "freva-rest", "web", "data-loader"],
        choices=["web", "db", "freva-rest", "data-loader"],
        help="The services to be deployed.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        help="Output directory where the manifests should be located.",
        default=None,
    )
    parser.add_argument(
        "--no-plugins",
        action="store_true",
        help="Do not setup core library to use plugins.",
    )
    parser.set_defaults(cli=create_manifest)



"""Simple script to start and stop the storage service."""

import json
import logging
import subprocess
import time
import urllib.request

# Set up logging
logging.basicConfig(
    format="%(name)s - %(levelname)s - %(message)s",
    datefmt="[%X]",
    level=logging.INFO,
)
logger = logging.getLogger("container-check")


def start_container(container_name: str) -> subprocess.Popen:
    """Start the container."""
    return subprocess.Popen(
        [
            "docker",
            "run",
            "--net=host",
            "-e",
            "ROOT_PW=foo",
            "-e",
            "VERSION=0.0.0",
            "-p",
            "5002:5002",
            container_name,
        ],
    )


def _check_container(process: subprocess.Popen) -> None:
    """Check if the container starts up."""
    try:
        if process.poll() is not None:
            raise RuntimeError("Container died.")
        res = urllib.request.Request(
            "http://localhost:5002/vault/status",
        )
        with urllib.request.urlopen(res) as response:
            if response.getcode() != 200:
                raise RuntimeError("Container not properly set up.")
            json_req = json.loads(response.read().decode())
    except Exception as error:
        logger.critical("The container failed: %s", error)
        raise RuntimeError(error) from None
    if json_req.get("status") != "unsealed":
        logger.critical("Vault did not unseal: %s", json_req.get("status", "500"))
        raise ValueError("Vault did not unseal")
    process.terminate()
    logger.info("Container seems to work!")


def check_container(container_name: str = "vault") -> None:
    """Check the status of the container."""
    proc = start_container(container_name)
    time.sleep(5)
    for _ in range(10):
        try:
            _check_container(proc)
            return
        except RuntimeError:
            time.sleep(5)
    proc.terminate()
    raise ValueError("Container deosn't seem to work.")


if __name__ == "__main__":
    check_container()



from __future__ import annotations

import curses
import json
import logging
import os
import re
from pathlib import Path
from typing import Any, Optional, cast

import npyscreen
import tomlkit
from npyscreen.wgwidget import Widget

from freva_deployment.utils import AD, asset_dir, config_file

logging.basicConfig(level=logging.DEBUG)
logger: logging.Logger = logging.getLogger("deploy-freva-tui")


class InfoMixin(Widget):
    """Mixin class to extend npyscreen widgets with an infobox."""

    def __init__(
        self,
        *args: Any,
        section: str = "",
        key: str = "",
        info: str = "",
        **kwargs: Any,
    ) -> None:
        name = kwargs.get("name", "Select")
        kwargs["name"] = f"{name}. Press Ctrl+F for more info."
        self.info = info or AD.get_config_info(section, key)
        super().__init__(*args, **kwargs)

    def edit(self) -> None:
        self.parent.current_info = self.info
        super().edit()


class TextInfo(InfoMixin, npyscreen.TitleText):
    """Extend the TitleText widget by an infobox."""


class DictInfo(npyscreen.TitleText):
    """Extend the TitleText widget by an infobox."""

    def __init__(
        self,
        *args: Any,
        section: str = "",
        key: str = "",
        info: str = "",
        value: Optional[dict[str, list[str]]] = None,
        **kwargs: Any,
    ) -> None:
        name = kwargs.get("name", "Select")
        kwargs["name"] = f"{name}. Press Ctrl+F for more info."
        self.info = info or AD.get_config_info(section, key)
        try:
            value_str = tomlkit.dumps(value or {}).strip().replace('"', "")
        except Exception:
            value_str = ""
        super().__init__(
            *args,
            value=value_str,
            **kwargs,
        )

    def edit(self) -> None:
        self.parent.current_info = self.info
        super().edit()

    @property
    def dict_value(self) -> dict[str, list[str]]:
        result: dict[str, list[str]] = {}
        # Replace colons with '=' to mimic TOML inline table
        user_input = super().value
        cleaned = re.sub(r"(\w+)\s*:", r"\1 =", user_input)
        # Quote bare words (values like aa, bb)
        cleaned = (
            "{"
            + (
                re.sub(
                    r'(?<!["\'])\b([a-zA-Z_][a-zA-Z0-9_-]*)\b(?!["\'])',
                    r'"\1"',
                    cleaned,
                )
                .lstrip("{")
                .rstrip("}")
            )
            + "}"
        )
        try:
            toml_res: dict[str, Any] = tomlkit.loads(f"result = {cleaned}")
        except Exception:
            toml_res = {}
        for key, values in (toml_res.get("result") or {}).items():
            if isinstance(values, str):
                result[key] = [values]
            else:
                result[key] = values

        return result


class FileInfo(InfoMixin, npyscreen.TitleFilename):
    """Extend the TitleFilename widget by an infobox."""


class ComboInfo(InfoMixin, npyscreen.TitleCombo):
    """Extend the TitleCombo widget by an infobox."""


class PasswordInfo(InfoMixin, npyscreen.TitlePassword):
    """Extend the TitlePassword widget by an infobox."""


class CheckboxInfo(InfoMixin, npyscreen.RoundCheckBox):
    """Extend the TitlePassword widget by an infobox."""


class FileSelector(npyscreen.FileSelector):
    """FileSelector widget that allows for filtering file extensions."""

    file_extentions: list[str]
    """List of allowed file extensions."""
    value: str
    """The value of this file selector."""

    def __init__(
        self, *args, file_extentions: str | list[str] = [".toml"], **kwargs
    ) -> None:
        if isinstance(file_extentions, str):
            self.file_extentions = [file_extentions]
        else:
            self.file_extentions = file_extentions
        super().__init__(*args, **kwargs)
        self.how_exited_handers[npyscreen.wgwidget.EXITED_ESCAPE] = self.exit

    def exit(self) -> None:
        """Exit the dialogue."""
        self.wCommand.value = ""
        self.value = ""
        self.exit_editing()

    def update_grid(self) -> None:
        if self.value:
            self.value = os.path.expanduser(self.value)
        if not os.path.exists(self.value):
            self.value = os.getcwd()
        if os.path.isdir(self.value):
            working_dir = Path(self.value)
        else:
            working_dir = Path(self.value).parent
        self.wStatus1.value = working_dir
        file_list, dir_list = [], []
        if working_dir.parent != working_dir:
            dir_list.append("..")
        try:
            for fn in working_dir.glob("*"):
                rel_path = str(fn.relative_to(working_dir))
                if not rel_path.startswith("."):
                    if fn.is_dir():
                        dir_list.append(str(fn) + os.sep)
                    elif fn.suffix in self.file_extentions:
                        file_list.append(str(fn))
        except OSError:
            npyscreen.notify_wait(
                title="Error", message="Could not read specified directory."
            )
        # DOES NOT CURRENTLY WORK - EXCEPT FOR THE WORKING DIRECTORY.  REFACTOR.
        # sort Filelist
        file_list.sort(key=str.casefold)
        dir_list.sort(key=str.casefold)
        if self.sort_by_extension:
            file_list.sort(key=self.get_extension)
        self.wMain.set_grid_values_from_flat_list(
            dir_list + file_list, reset_cursor=False
        )
        self.display()


def selectFile(starting_value: str = "", *args, **keywords):
    F = FileSelector(*args, **keywords)
    F.set_colors()
    F.wCommand.show_bold = True
    if starting_value:
        if not os.path.exists(
            os.path.abspath(os.path.expanduser(starting_value))
        ):
            F.value = os.getcwd()
        else:
            F.value = starting_value
            F.wCommand.value = starting_value
    else:
        F.value = os.getcwd()
    F.update_grid()
    F.edit()
    return F.wCommand.value


class BaseForm(npyscreen.FormMultiPageWithMenus, npyscreen.FormWithMenus):
    """Base class for forms."""

    _num: int = 0
    input_fields: dict[str, tuple[npyscreen.TitleText, bool]] = {}
    """Dictionary of input fields: the key of the dictionary represents the name
       of the key in the in config toml input files. Values represent a tuple of
       npysceen types that display the input information on this key to the
       user and a boolean indicating whether or not this variable is mandatory.
    """
    certificates: list[str] = []
    """The type of certificate files this step needs."""

    playbook_dir: Path = asset_dir / "playbooks"
    """Default playbook location."""

    current_info: Optional[str] = None
    """The currently selected widget."""

    def get_config(self, key) -> dict[str, str | bool | list[str]]:
        """Read the configuration for a step."""
        try:
            cfg = self.parentApp.config[key].copy()
            if not isinstance(cfg, dict):
                cfg = {}
        except (KeyError, AttributeError, TypeError):
            cfg = {}
        return cfg

    @property
    def num(self) -> str:
        """Calculate the number for enumerations of any input field."""
        self._num += 1
        return f"{self._num}. "

    def check_config(
        self,
        notify: bool = True,
    ) -> dict[str, str | dict[str, str | list | int | bool | None]] | None:
        """Check if the from entries are valid."""
        config = {}
        for key, (obj, mandatory) in self.input_fields.items():
            if hasattr(obj, "dict_value"):
                value = obj.dict_value
            elif hasattr(obj, "values"):
                value = obj.values[obj.value]
            else:
                value = obj.value
            if isinstance(value, str):
                if not value and self.use.value and mandatory and notify:
                    msg = f"MISSING ENTRY FOR {self.step}: {obj.name}"
                    npyscreen.notify_confirm(msg, title="ERROR")
                    return None
            config[key] = value
        cfg = config.copy()
        cfg = config
        for key, value in cfg.items():
            if key in self.list_keys:
                cfg[key] = value.split(",")
        return cfg

    def draw_form(self) -> None:
        """Overload the draw_from method from, this is done to add menus."""
        super().draw_form()
        menus = [
            " " + self.__class__.MENU_KEY + ":Main Menu ",
            "^K:Prev. Tab ",
            "^L:Next Tab ",
            "^R:Run ",
            "^S:Save ",
            "^O:Load ",
            "^E:Exit ",
        ]
        y, x = self.display_menu_advert_at()
        for n, menu_advert in enumerate(menus):
            if isinstance(menu_advert, bytes):
                menu_advert = menu_advert.decode("utf-8", "replace")
            X = sum([len(menu) for menu in menus[:n]]) + x
            self.add_line(
                y,
                X,
                menu_advert,
                self.make_attributes_list(menu_advert, curses.A_NORMAL),
                self.columns - X - 1,
            )

    def previews_form(self, *args, **keywords) -> None:
        return self.change_forms(*args, reverse=True, **keywords)

    def change_forms(self, *args, reverse=False, **keywords) -> None:
        """Cycle between the deployment config forms."""
        for step in self.parentApp._steps_lookup.keys():
            if self.name.lower().startswith(step):
                name = step
                break
            elif "rest" in self.name.lower():
                name = "freva_rest"
            elif self.name.lower().startswith("database"):
                name = "db"
                break
        keys = self.parentApp._steps_lookup.keys()
        steps = list(self.parentApp._steps_lookup.values())
        if reverse:
            change_to = dict(zip(keys, [steps[-1]] + steps[:-1]))
        else:
            change_to = dict(zip(keys, steps[1:] + [steps[0]]))
        self.parentApp.current_form = name
        self.parentApp.change_form(change_to[name])

    def run_deployment(self, *args) -> None:
        """Switch to the deployment setup form."""
        self.parentApp.current_form = self.name
        self.parentApp.change_form("SETUP")

    def add_variables(self):
        """Add some environment variables for running the deployment."""
        popup = VarForm()
        popup.edit()
        del popup

    def clear_cache(self):
        """Clear the app cache."""
        with open(self.parentApp.cache_dir / "freva_deployment.json", "w") as f:
            json.dump({}, f, indent=3)
        self.parentApp.reset()

    def setTheme(self, theme: str) -> None:
        self.theme_manager.theme = getattr(npyscreen.Themes, theme)
        self.theme_manager.theme = theme  # getattr(npyscreen.Themes, theme)
        self.parentApp.DEFAULT_THEME = theme
        self.create()
        self.display()
        self.display()

    def show_info(self, *args: Any, **kwargs: Any) -> None:
        """Display an info if present."""
        if isinstance(self.current_info, str):
            npyscreen.notify_confirm(
                self.current_info, title="Detailed Information"
            )

    def create(self) -> None:
        """Setup the form."""
        self.how_exited_handers[npyscreen.wgwidget.EXITED_ESCAPE] = (
            self.parentApp.exit_application
        )
        self.add_handlers({"^F": self.show_info})
        self.add_handlers({"^O": self.parentApp.load_dialog})
        self.add_handlers({"^S": self.parentApp.save_dialog})
        self.add_handlers({"^K": self.previews_form})
        self.add_handlers({"^L": self.change_forms})
        self.add_handlers({"^R": self.run_deployment})
        self.add_handlers({"^E": self.parentApp.exit_application})
        # The menus are created here.
        self.menu = self.add_menu(name="Main Menu", shortcut="^M")
        self.submenu = self.menu.addNewSubmenu("Deployment Menu")
        self._change_form = self.parentApp.change_form
        self.menu.addItemsFromList(
            [
                ("Set Variables", self.add_variables, "^V"),
                ("Save Config", self.parentApp.save_dialog, "^S"),
                ("Load Config", self.parentApp.load_dialog, "^L"),
                ("Run Deployment", self.run_deployment, "^R"),
                ("Clear App Cache", self.clear_cache, "^D"),
                ("Exit Application", self.parentApp.exit_application, "^E"),
            ]
        )
        self.submenu.addItemsFromList(
            [
                ("Core Deployment", self._change_form, "c", "c", ("MAIN",)),
                ("Web Deployment", self._change_form, "w", "w", ("SECOND",)),
                ("DB Deployment", self._change_form, "d", "d", ("THIRD",)),
                (
                    "Freva-Rest Deployment",
                    self._change_form,
                    "s",
                    "s",
                    ("FOURTH",),
                ),
                ("Run Deployment", self.run_deployment, "^R"),
            ]
        )
        self.use = self.add(
            npyscreen.CheckBox,
            max_height=2,
            value=self.step in self.parentApp._steps,
            editable=True,
            name="Check to set up the {}".format(self.step),
            scroll_exit=True,
        )
        self._add_widgets()


class VarForm(npyscreen.FormMultiPageActionWithMenus):
    """Definition of the form that applies the actual deployment."""

    _num: int = 0

    @property
    def num(self) -> str:
        """Calculate the number for enumerations of any input field."""
        self._num += 1
        return f"{self._num}. "

    def create(self):
        npyscreen.setTheme(npyscreen.Themes.ElegantTheme)
        self.cfg = None
        self.keys = []
        self.values = []
        self.cfg = tomlkit.loads(config_file.read_text())
        for key, value in self.cfg["variables"].items():
            self.add_widget_intelligent(
                npyscreen.TitleText, name=f"{self.num}{key}:", value=value
            )
        self.add_widget_intelligent(
            npyscreen.TitleText,
            name="__________________________",
            editable=False,
        )
        self.delete = self.add_widget_intelligent(
            npyscreen.TitleCombo,
            name="Delete Keys",
            value=0,
            relx=4,
            values=[""] + list(self.cfg["variables"].keys()),
        )
        self.add_btn = self.add_widget_intelligent(
            npyscreen.ButtonPress,
            name="Press for new key",
            use_max_space=True,
            when_pressed_function=self._add_new_keys,
        )
        self.add_widget_intelligent(
            npyscreen.TitleText,
            name="__________________________",
            editable=False,
        )

    def _add_new_keys(self, *args, **kwargs) -> None:
        buttons = [self._widgets__.pop(-1), self._widgets__.pop(-1)]  # type: ignore
        self.keys.append(
            self.add_widget_intelligent(
                npyscreen.TitleText,
                name=f"Set a new key:",
                value="",
                editable=True,
            )
        )
        self.values.append(
            self.add_widget_intelligent(
                npyscreen.TitleText,
                name=f"Set a the according value:",
                value="",
                editable=True,
            )
        )

        self._widgets__ = cast(list[Any], self._widgets__ + buttons)  # type: ignore
        self.display()

    def on_ok(self) -> None:
        if self.delete.value > 0:
            key = list(self.cfg["variables"].keys())[self.delete.value - 1]
            del self.cfg["variables"][key]
        for key, value in zip(self.keys, self.values):
            if key.value and value.value:
                self.cfg["variables"][key.value.upper()] = value.value

        toml_str = tomlkit.dumps(self.cfg)
        self.value = self.key = None
        with config_file.open("w", encoding="utf-8") as f_obj:
            f_obj.write(toml_str)
        self.cfg = None

    def on_cancel(self) -> None:
        self.cfg = None



"""Definition of a custom ansible stdout callback."""

from __future__ import absolute_import, division, print_function

from typing import IO, Optional

__metaclass__ = type

DOCUMENTATION = """
    author: DRKZ CLINT
    name: freva_deployment
    type: stdout
    short_description: YAML-ized Ansible screen output and write to log
    description:
        - Ansible output that can be quite a bit easier to read than the
          default JSON formatting.
        - Parse output to a log file at the same time.
    options:
        log_file:
            default: temporary file
            type: str
            description:
                - Set the path to the log file.
            env:
                - name: ANSIBLE_LOG_PATH
    extends_documentation_fragment:
      - default_callback
    requirements:
      - set as stdout in configuration
"""
import json
import os
from tempfile import NamedTemporaryFile
from typing import Optional

from ansible.constants import MODULE_NO_JSON
from ansible.executor.task_result import TaskResult
from ansible.plugins.callback import (
    module_response_deepcopy,
    strip_internal_keys,
)

# -----------------------------------------------------------------------------
# Select a YAML-style base callback that works with multiple Ansible versions
# -----------------------------------------------------------------------------
try:
    # Preferred: built-in yaml callback (ansible-core)
    from ansible.plugins.callback.yaml import (
        CallbackModule as BaseYamlCallback,
    )
except Exception:
    try:
        # Fallback: old community.general yaml callback
        from ansible_collections.community.general.plugins.callback.yaml import (  # noqa: E501
            CallbackModule as BaseYamlCallback,
        )
    except Exception:
        # Last resort: default stdout callback (no YAML formatting, but works)
        from ansible.plugins.callback.default import (
            CallbackModule as BaseYamlCallback,
        )


class CallbackModule(BaseYamlCallback):
    """
    Custom Ansible callback plugin that logs output to a file in single-line JSON format
    and displays output to stdout in YAML format.

    Attributes
    ----------
    CALLBACK_VERSION : float
        Version of the callback plugin.
    CALLBACK_TYPE : str
        Type of the callback plugin.
    CALLBACK_NAME : str
        Name of the callback plugin.
    log_file : IO
        File object for the log file.
    """

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = "stdout"
    CALLBACK_NAME = "deployment_plugin"

    def __init__(self) -> None:
        """
        Initializes the callback plugin.

        Parameters
        ----------
        display : Optional[Display], optional
            Ansible display object for managing console output, by default None.
        """
        log_file = (
            os.getenv("DEPLOYMENT_LOG_PATH")
            or NamedTemporaryFile(suffix=".log", delete=False).name
        )
        self.log_file: IO[str] = open(log_file, "w")
        super().__init__()
        self._plugin_options["result_format"] = "yaml"
        self._plugin_options["pretty_results"] = True

    def log_result(self, result: TaskResult) -> None:
        """
        Logs the result as a single-line JSON string to the log file.

        Parameters
        ----------
        result : TaskResult
            Ansible task result object.
        """
        dump = {
            "task": result.task_name,
            "result": strip_internal_keys(
                module_response_deepcopy(result._result)
            ),
        }
        single_line_json = json.dumps(dump).replace("\n", "").replace("\r", "")
        self.log_file.write(single_line_json + "\n")
        self.log_file.flush()

    def v2_runner_on_ok(self, result: TaskResult) -> None:
        if (
            result._task.action not in MODULE_NO_JSON
            or "ansible_job_id" not in result._result
        ):
            self.log_result(result)
        super().v2_runner_on_ok(result)

    def v2_runner_on_async_ok(self, result: TaskResult) -> None:
        if (
            result._task.action not in MODULE_NO_JSON
            or "ansible_job_id" not in result._result
        ):
            self.log_result(result)
        super().v2_runner_on_async_ok(result)

    def __del__(self) -> None:
        """
        Closes the log file when the callback plugin is destroyed.
        """
        self.log_file.close()



"""Generate keys."""

import datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

_CRYPTO = True
try:
    from cryptography import x509
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import hashes, serialization
    from cryptography.hazmat.primitives.asymmetric import (
        ec,
        ed25519,
        padding,
        rsa,
    )
    from cryptography.x509.oid import NameOID
except ImportError:
    _CRYPTO = False


class RandomKeys:
    """Generate public and private server keys.

    Parameters:
        base_name (str): The path prefix for all key files.
        common_name (str): The common name for the certificate.
    """

    def __init__(
        self, base_name: str = "freva", common_name: str = "localhost"
    ) -> None:
        self.base_name = base_name
        self.common_name = common_name or "localhost"
        self._private_key_pem: Optional[bytes] = None
        self._public_key_pem: Optional[bytes] = None
        self._private_key: Optional["rsa.RSAPrivateKey"] = None
        self._temp_dir = TemporaryDirectory("random_keys")

    @staticmethod
    def _check_crypto() -> None:
        if not _CRYPTO:
            raise ImportError(
                "Please install the `cryptography` python module."
                " in order to generate certificates."
            )

    def check_cert_key_pair(self, cert_path: str, key_path: str) -> bool:
        """Validate a certificate/key pair by checking:
            - The certificate is currently valid (not expired).
            - The certificate and private key match.
        Parameters
        ----------
        cert_path : str
            Path to the PEM-encoded certificate file.
        key_path : str
            Path to the PEM-encoded private key file.
        allow_self_signed : bool, optional
            Whether to consider self-signed certificates as valid, by default False.

        Returns
        -------
        bool
            True if the certificate/key pair is valid, matching, and trusted
            False otherwise.
        """
        self._check_crypto()
        try:
            with open(cert_path, "rb") as f:
                cert_data = f.read()
            cert = x509.load_pem_x509_certificate(cert_data, default_backend())

            with open(key_path, "rb") as f:
                key_data = f.read()
            private_key = serialization.load_pem_private_key(
                key_data, password=None, backend=default_backend()
            )
            public_key = cert.public_key()
        except Exception:
            return False

        now = datetime.datetime.now(datetime.timezone.utc)
        if not (cert.not_valid_before_utc <= now <= cert.not_valid_after_utc):
            return False
        if cert.issuer == cert.subject:
            return False

        message = b"test"
        try:
            if isinstance(private_key, rsa.RSAPrivateKey) and isinstance(
                public_key, rsa.RSAPublicKey
            ):
                signature = private_key.sign(
                    message, padding.PKCS1v15(), hashes.SHA256()
                )
                public_key.verify(
                    signature, message, padding.PKCS1v15(), hashes.SHA256()
                )
                return True

            elif isinstance(
                private_key, ec.EllipticCurvePrivateKey
            ) and isinstance(public_key, ec.EllipticCurvePublicKey):
                signature = private_key.sign(message, ec.ECDSA(hashes.SHA256()))
                public_key.verify(signature, message, ec.ECDSA(hashes.SHA256()))
                return True

            elif isinstance(
                private_key, ed25519.Ed25519PrivateKey
            ) and isinstance(public_key, ed25519.Ed25519PublicKey):
                signature = private_key.sign(message)
                public_key.verify(signature, message)
                return True

            else:
                return False

        except Exception:
            return False

    @property
    def private_key(self) -> "rsa.RSAPrivateKey":
        if self._private_key is not None:
            return self._private_key
        self._private_key = rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )
        return self._private_key

    @property
    def private_key_pem(self) -> bytes:
        """Create a new private key pem if it doesn't exist."""
        self._check_crypto()
        if self._private_key_pem is None:
            self._private_key_pem = self.private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
        return self._private_key_pem

    @property
    def public_key_pem(self) -> bytes:
        """
        Generate a public key pair using RSA algorithm.

        Returns:
            bytes: The public key (PEM format).
        """
        self._check_crypto()
        if self._public_key_pem is None:
            public_key = self.private_key.public_key()
            self._public_key_pem = public_key.public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
        return self._public_key_pem

    @property
    def private_key_file(self) -> str:
        """
        Save the private key to a file.

        Returns:
            str: The filename of the private key file.
        """
        self._check_crypto()
        temp_file = (Path(self._temp_dir.name) / self.base_name).with_suffix(
            ".key"
        )
        if not temp_file.is_file():
            temp_file.write_bytes(self.private_key_pem)
        return str(temp_file)

    @property
    def public_key_file(self) -> str:
        """
        Save the public key to a file.

        Returns:
            str: The filename of the public key file.
        """
        temp_file = (Path(self._temp_dir.name) / self.base_name).with_suffix(
            ".pem"
        )
        if not temp_file.is_file():
            temp_file.write_bytes(self.public_key_pem)
        return str(temp_file)

    @property
    def public_chain_file(self) -> str:
        """
        Save the public key and certificate chain to a file.

        Returns:
            str: The filename of the certificate chain file.
        """
        temp_file = (Path(self._temp_dir.name) / self.base_name).with_suffix(
            ".crt"
        )
        if not temp_file.is_file():
            temp_file.write_bytes(self.certificate_chain)
        return str(temp_file)

    def create_self_signed_cert(self) -> "x509.Certificate":
        """
        Create a self-signed certificate using the public key.

        Returns
        -------
            x509.Certificate: The self-signed certificate.
        """
        self._check_crypto()
        csr = (
            x509.CertificateSigningRequestBuilder()
            .subject_name(
                x509.Name(
                    [x509.NameAttribute(NameOID.COMMON_NAME, self.common_name)]
                )
            )
            .sign(self.private_key, hashes.SHA256(), default_backend())
        )
        # Add SANs
        san = x509.SubjectAlternativeName(
            [
                x509.DNSName(f"{self.common_name}"),
                x509.DNSName(f"www.{self.common_name}"),
                x509.DNSName("web-reverse-proxy"),
                x509.DNSName("www.web-reverse-proxy"),
            ]
        )
        certificate = (
            x509.CertificateBuilder()
            .subject_name(csr.subject)
            .issuer_name(csr.subject)
            .add_extension(san, critical=False)
            .public_key(csr.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.datetime.now(datetime.timezone.utc))
            .not_valid_after(
                datetime.datetime.now(datetime.timezone.utc)
                + datetime.timedelta(days=10 * 365)
            )
            .sign(self.private_key, hashes.SHA256(), default_backend())
        )

        return certificate

    @property
    def certificate_chain(self) -> bytes:
        """The certificate chain."""
        certificate = self.create_self_signed_cert()
        certificate_pem = certificate.public_bytes(serialization.Encoding.PEM)
        return self.public_key_pem + certificate_pem


if __name__ == "__main__":
    # Create an instance of RandomKeys
    keys = RandomKeys()

    # Output the public key
    print("Public Key:")
    print(Path(keys.private_key_file).read_text())
    # Output the certificate chain
    print("\nPublic Key:")
    print(Path(keys.public_chain_file).read_text())



"""Collection of utils for deployment."""

from __future__ import annotations

import hashlib
import ipaddress
import json
import os
import re
import shutil
import socket
import sys
import sysconfig
from pathlib import Path
from typing import Any, Dict, MutableMapping, NamedTuple, Optional, Union, cast
from urllib.request import urlopen

import appdirs
import namegenerator
import tomlkit
from rich.console import Console
from rich.prompt import Prompt

from .error import ConfigurationError
from .keys import RandomKeys
from .logger import logger

RichConsole = Console(markup=True, force_terminal=True)

config_text = """
### This is the global config file for the freva deployment
[general]
[variables]
### you can set here different *global* variables that are
### evaluated in the deployment configurations. For example
### if you set here the variable USER = "foo" then you can
### use the this defined variable in the inventory file to
### set the ansible user: anisble_user="${USER}"

# USER="myusername"
# USER_GROUP = "myusergroup"
"""

password_prompt = (
    "[green]Choose[/] a [b]master password[/], this password will be used to:\n"
    "- create the [magenta]mysql root[/] password\n"
    "- set the [magenta]django admin[/] web password\n"
    "- set the [magenta]mongo db[/] user password\n"
    "[b]Note:[/] Ideally this password can be shared amongst other [i]admins[/].\n"
    "[b][green]choose[/] master password[/]"
)

ServiceInfo = NamedTuple(
    "ServiceInfo", [("name", str), ("python_path", str), ("hosts", str)]
)


class AssetDir:
    this_module = "freva_deployment"

    def __init__(self):
        self._inventory_content = None
        self._user_asset_dir = (
            Path(appdirs.user_data_dir()) / "freva" / "deployment"
        )
        self._user_config_dir = (
            Path(appdirs.user_config_dir()) / "freva" / "deployment"
        )

    @staticmethod
    def is_bundeled() -> bool:
        """Check if we are running a bundled version of the code."""
        if getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
            return True
        return False

    @staticmethod
    def get_dirs(user: bool = True, key: str = "data") -> Path:
        """Get the 'scripts' and 'purelib' directories we'll install into.

        This is now a thin wrapper around sysconfig.get_paths(). It's not inlined,
        because some tests mock it out to install to a different location.
        """

        if user:
            if (sys.platform == "darwin") and sysconfig.get_config_var(
                "PYTHONFRAMEWORK"
            ):
                return Path(sysconfig.get_paths("osx_framework_user")[key])
            return Path(sysconfig.get_paths(os.name + "_user")[key])
        # The default scheme is 'posix_prefix' or 'nt', and should work for e.g.
        # installing into a virtualenv
        return Path(sysconfig.get_paths()[key])

    @property
    def asset_dir(self) -> Path:
        if self.is_bundeled():
            return Path(__file__).parent / "assets"

        data_dir = self.get_dirs(False) / "share" / "freva" / "deployment"
        user_dir = self.get_dirs(True) / "share" / "freva" / "deployment"
        data_dir_local = (
            self.get_dirs(False) / "local" / "share" / "freva" / "deployment"
        )
        for path in (data_dir, user_dir, data_dir_local):
            if path.is_dir():
                return path
        raise ConfigurationError(
            "Could not find asset dir, please consider reinstalling the package."
        )

    @property
    def inventory_file(self) -> Path:
        return self.asset_dir / "config" / "inventory.toml"

    @property
    def inventory_content(self) -> str:
        if self._inventory_content is None:
            self._inventory_content = self.inventory_file.read_text()
        return self._inventory_content

    def get_config_info(self, section: str, key: str) -> str:
        """Get the information of a given config key in the config."""
        if section not in self.inventory_content:
            raise ValueError("SectionNotFound")
        content = self.inventory_content.split(f"[{section}]")[-1].splitlines()
        key_pos = None
        # Find the line number of the specified key
        for i, line in enumerate(content):
            if line.strip().startswith(key):
                key_pos = i
                break
        if key_pos is None:
            return ""
        # Collect the explanation block directly above the specified key
        explanation = []
        for i in range(key_pos - 1, -1, -1):
            if content[i].strip().startswith("#"):
                explanation.append(
                    content[i].strip().replace("##", "#").strip("#").strip()
                )
            else:
                break

        return " ".join(reversed(explanation)).strip()

    @property
    def config_dir(self):
        inventory_file = self._user_config_dir / "inventory.toml"
        if inventory_file.exists():
            return self.asset_dir
        inventory_file.parent.mkdir(exist_ok=True, parents=True)
        try:
            inventory_file.unlink()
        except FileNotFoundError:
            pass
        shutil.copy2(self.asset_dir / "config" / "inventory.toml", inventory_file)
        return self._user_config_dir

    @property
    def config_file(self):
        if os.access(self.config_dir, os.W_OK):
            cfg_file = self.config_dir / "freva-deployment.config"
        else:
            cfg_file = self._user_config_dir / "freva-deployment.config"
        if not cfg_file.exists():
            cfg_file.parent.mkdir(exist_ok=True, parents=True)
            with cfg_file.open("w", encoding="utf-8") as f_obj:
                f_obj.write(config_text)
        return cfg_file


AD = AssetDir()
asset_dir = AD.asset_dir
config_dir = AD.config_dir
config_file = AD.config_file
is_bundeled = AD.is_bundeled()


def is_localhost(host: str) -> bool:
    """Check if a given hostname is localhost."""
    try:
        ip = ipaddress.ip_address(host)
    except ValueError:
        try:
            ip = ipaddress.ip_address(socket.gethostbyname(host))
        except socket.gaierror:
            return False

    local_addrs = set()
    for addr_info in socket.getaddrinfo(socket.gethostname(), None):
        local_addrs.add(addr_info[4][0])

    # Add standard localhost addresses
    local_addrs.update({"127.0.0.1", "::1"})
    try:
        for iface_info in socket.getaddrinfo(socket.getfqdn(), None):
            local_addrs.add(iface_info[4][0])
    except socket.gaierror:
        pass
    return str(ip) in local_addrs


def get_cache_information(
    redis_host: Optional[str] = None,
    redis_port: Optional[str] = None,
    scheduler_host: Optional[str] = None,
    scheduler_port: Optional[str] = None,
) -> Dict[str, str]:
    """Create all information we need to setup the redis cache and the data portal."""
    user = ssl_cert = ssl_key = ""
    if redis_host:
        keys = RandomKeys(
            common_name=redis_host.rpartition("//")[-1].partition(":")[0]
        )
        user = namegenerator.gen()
        ssl_cert = keys.certificate_chain.decode("utf-8")
        ssl_key = keys.private_key_pem.decode("utf-8")

    if scheduler_host and scheduler_port:
        scheduler_host = f"{scheduler_host}:{scheduler_port}"
    if redis_host and redis_port:
        redis_host = f"redis://{redis_host}:{redis_port}"
    elif redis_host and not redis_port:
        redis_host = f"redis://{redis_host}"
    return {
        "ssl_cert": ssl_cert,
        "ssl_key": ssl_key,
        "host": redis_host or "",
        "user": user,
        "passwd": "",
        "scheduler_host": scheduler_host or "",
        "cache_exp": "86400",
    }


def get_current_file_dir(inp_dir: str | Path, value: str) -> str:
    """Get a path with the CFD as a variable."""
    if "${CFD}" in value.upper() or "$CFD" in value.upper():
        value = value.replace("${CFD}", "$CFD")
        part_1, _, part_2 = value.partition("$CFD/")
        return str(Path(inp_dir, *Path(part_2).parts))
    return value


def _convert_dict(
    inp_dict: dict[str, str | dict[str, Any]],
    variables: dict[str, str],
    cfd: Path,
) -> None:
    for key, value in inp_dict.items():
        if isinstance(value, dict):
            _convert_dict(value, variables, cfd)
        elif isinstance(value, str):
            for warn, variable in variables.items():
                if f"${{{warn}}}" in value or f"${warn}" in value:
                    value = value.replace(f"${{{warn}}}", f"${warn}")
                    value = value.replace(f"${warn}", variable)
            inp_dict[key] = get_current_file_dir(cfd, value)


def _update_config(
    new_config: MutableMapping[str, Any], old_config: MutableMapping[str, Any]
) -> None:
    for key, value in old_config.items():
        if key in new_config and isinstance(value, dict):
            _update_config(new_config[key], old_config[key])
        else:
            new_config[key] = value


def _create_new_config(inp_file: Path) -> Path:
    """Update any old configuration file to a newer version."""
    config = cast(
        Dict[str, Dict[str, Dict[str, Union[str, float, int, bool]]]],
        tomlkit.loads(inp_file.read_text()),
    )
    keys_to_check = {
        "freva_rest": [
            "freva_rest_host",
            "search_server_host",
            "mongodb_server_host",
            "redis_host",
            "oidc_url",
            "oidc_client",
            "oidc_client_secret",
            "oidc_token_claims",
            "data_loader_portal_hosts",
            "admin_user",
        ],
        "web": [
            "web_host",
            "admin_user",
            "chatbot_host",
        ],
        "db": ["vault_host", "db_host", "admin_user"],
        "core": ["core_host"],
    }
    create_backup = False
    config_tmpl = cast(
        Dict[str, Any], tomlkit.loads(AD.inventory_file.read_text())
    )
    # Legacy keys:
    if "solr" in config or "databrowser" in config:
        create_backup = True
        if "solr" in config:
            config["freva_rest"] = config.pop("solr")
            for key in ("port", "mem"):
                if key in config["freva_rest"]["config"]:
                    config["freva_rest"]["config"][f"solr_{key}"] = config[
                        "freva_rest"
                    ]["config"].pop(key)
        else:
            config["freva_rest"] = config.pop("databrowser")
            config["freva_rest"]["config"]["freva_rest_port"] = config[
                "freva_rest"
            ]["config"].pop("databrowser_port", "")

    for section, keys in keys_to_check.items():
        if isinstance(config[section].get("config"), dict):
            create_backup = True
            config[section]["config"][f"{section}_host"] = str(
                config[section]["hosts"]
            )
            config[section] = config[section]["config"]  # type: ignore
        config[section].pop(f"{section}_playbook", "")
        for key in keys:
            if key not in tomlkit.dumps(config):
                config[section][key] = config_tmpl[section][key]
                create_backup = True
    for key in ("deployment_method", "kubernetes"):
        if key not in config:
            create_backup = True
            config[key] = config_tmpl[key]
    for key in config_tmpl["kubernetes"]:
        if key not in config["kubernetes"]:
            create_backup = True
            config["kubernetes"][key] = config_tmpl["kubernetes"][key]
    if create_backup:
        backup_file = inp_file.with_suffix(inp_file.suffix + ".bck")
        logger.info(
            "Updating config file, original config can be found in %s",
            backup_file,
        )
        backup_file.write_text(inp_file.read_text())
        _update_config(config_tmpl, config)
        inp_file.write_text(tomlkit.dumps(config_tmpl))
    return inp_file


def load_config(inp_file: str | Path, convert: bool = False) -> dict[str, Any]:
    """Load the inventory toml file and replace all environment variables."""
    inp_file = _create_new_config(Path(inp_file).expanduser().absolute())
    variables = cast(
        dict[str, str], tomlkit.loads(config_file.read_text())["variables"]
    )
    config = tomlkit.loads(inp_file.read_text(encoding="utf-8"))
    if convert:
        _convert_dict(config, variables, inp_file.parent)
    return config


def get_setup_for_service(
    service: str, setups: list[ServiceInfo]
) -> tuple[str, str]:
    """Get the setup of a service configuration."""
    for setup in setups:
        if setup.name == service:
            return setup.python_path, setup.hosts
    raise ConfigurationError("Service not found")


def read_db_credentials(db_host: str, port: int = 5002) -> dict[str, str]:
    """Read database config."""
    sha = hashlib.sha512("".encode()).hexdigest()
    url = f"http://{db_host}:{port}/vault/data/{sha}"
    with urlopen(url) as res:
        return json.loads(res.read().decode())


def get_email_credentials() -> tuple[str, str]:
    """Read login credentials for email server.

    Returns
    =======
    tuple: username and password
    """
    username: str = os.environ.get("EMAIL_USER", "") or ""
    password: str = os.environ.get("EMAIL_PASSWD", "") or ""
    msg = (
        "\nThe web will need login credentials to connect to the [b green]mail server [/]"
        "that has been set up.\nYou should now enter your [it]login credentials[/].\n"
        "[b]Note:[/]These credentials will be securely stored in an encrypted vault\n"
    )
    if not username or not password:
        RichConsole.print(msg)
        if not username:
            username = Prompt.ask("[green b]Username[/] for mail server")
        if not password:
            password = Prompt.ask(
                "[green b]Password[/] for mail server", password=True
            )
    return username, password


def get_passwd(master_pass: Optional[str] = None, min_characters: int = 8) -> str:
    """Create a secure password.

    Parameters
    ==========
    min_characters:
        The minimum length of the password (default 8)

    Returns
    =======
    str: The password
    """
    logger.debug("Creating Password")
    msg = ""
    while True:
        try:
            if master_pass:
                return _passwd_is_good(master_pass, min_characters)
            return _create_passwd(min_characters, msg)
        except ValueError as e:
            RichConsole.print(f"[red]{e}[/]")
            msg = "[b red]re-enter[/] master password"
            master_pass = ""


def _passwd_is_good(master_pass: str, min_characters: int) -> str:
    is_ok: bool = len(master_pass) > min_characters
    for check in ("[a-z]", "[A-Z]", "[0-9]"):
        if not re.search(check, master_pass):
            is_ok = False
            break
    forbidden_characters = ":/?#[]@%"
    is_ok = is_ok and len([True for c in "[_$^&*-!]" if c in master_pass]) > 0
    for character in forbidden_characters:
        if character in master_pass:
            is_ok = False
            break
    if not is_ok:
        raise ValueError(
            (
                "Password must confirm the following constraints:\n"
                f"- {min_characters} alphanumeric characters long,\n"
                "- have both, lower and upper case characters,\n"
                "- have at least one special special character.\n"
                f"- must *not* not contain: {forbidden_characters}"
            )
        )
    return master_pass


def _create_passwd(min_characters: int, msg: str = "") -> str:
    """Create passwords."""
    master_pass = Prompt.ask(msg or password_prompt, password=True)
    _passwd_is_good(master_pass, min_characters)
    master_pass_2 = Prompt.ask(
        "[bold green]re-enter[/] master password", password=True
    )
    if master_pass != master_pass_2:
        raise ValueError("Passwords do not match")
    return master_pass



import argparse
from urllib.request import urlretrieve

__version__ = "2601.0.0"

FREVA_PYTHON_VERSION = "3.13"
AVAILABLE_PYTHON_VERSIONS = ["3.9", "3.10", "3.11", "3.12", "3.13"]
AVAILABLE_CONDA_ARCHS = [
    "linux-64",
    "linux-aarch64",
    "linux-ppc64le",
    "linux-s390x",
    "osx-64",
    "osx-arm64",
]

AUX_URL = (
    "https://raw.githubusercontent.com/freva-org/freva-legacy/main/assets/"
    "evaluation_system.conf"
)


def download_auxiliry_data():
    """Download any data that needs to be downloaded."""

    def reporthook(count, block_size, total_size):
        if count == 0:
            return
        frac = count * block_size / total_size
        percent = int(100 * frac)
        bar = "#" * int(frac * 40)
        msg = "Downloading: [{0:<{1}}] | {2}% Completed".format(
            bar, 40, round(percent, 2)
        )
        print(msg, end="\r", flush=True)
        if frac >= 1:
            print()

    urls = {
        AUX_URL: (
            "assets/share/freva/deployment/config/evaluation_system.conf.tmpl"
        )
    }

    for source, target in urls.items():
        urlretrieve(source, filename=target, reporthook=reporthook)


if __name__ == "__main__":
    app = argparse.ArgumentParser()
    app.add_argument("-v", "--version", action="store_true", default=False)
    args = app.parse_args()
    if args.version:
        print(__version__)
    else:
        download_auxiliry_data()



"""Definition of the logger."""

import logging

from rich.console import Console
from rich.logging import RichHandler

logger_stream_handle = RichHandler(
    rich_tracebacks=True,
    show_path=False,
    console=Console(soft_wrap=True, stderr=True),
)
logging.basicConfig(
    format="%(name)s - %(message)s",
    level=logging.INFO,
    handlers=[logger_stream_handle],
    datefmt="[%X]",
)
logger = logging.getLogger("freva-deployment")


def set_log_level(verbosity: int) -> None:
    """Set the log level of the logger."""
    logger.setLevel(max(logging.INFO - 10 * verbosity, logging.DEBUG))



"""Ansible runner interaction."""

import atexit
import json
import os
import sys
from copy import deepcopy
from getpass import getuser
from multiprocessing import get_context
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Any, Dict, List, Optional, Union

import paramiko
import yaml
from rich import print as pprint
from rich.console import Console
from rich.live import Live
from rich.spinner import Spinner

from .error import DeploymentError
from .logger import logger
from .utils import is_bundeled


def _del_path(inp_path: Path) -> None:
    tmp_path = inp_path.with_suffix(".cfg.tmp")
    if inp_path.is_file():
        inp_path.unlink()
    if tmp_path.is_file():
        inp_path.write_text(tmp_path.read_text())
        tmp_path.unlink()


class SubProcess:
    """Class that holds the exitcode and the stdout of a process."""

    def __init__(
        self, exitcode: Optional[int], stdout: str = "", log: str = ""
    ) -> None:
        self.returncode = exitcode or 0
        self.stdout = stdout
        self.log = log

    @classmethod
    def run_ansible_playbook(
        cls,
        cwd: Path,
        command: List[str],
    ) -> None:
        from ansible.cli.playbook import main

        current_dir = os.path.abspath(os.curdir)
        try:
            os.chdir(cwd)
            main(command)
        finally:
            os.chdir(current_dir)


def run_command(
    cwd: str,
    command: List[str],
    env: Optional[Dict[str, str]] = None,
    capture_output: bool = False,
):
    os_env = deepcopy(os.environ)
    env = env or {}
    stdout = sys.stdout
    with TemporaryDirectory(prefix="AnsibleRunner") as temp_dir:
        logger_file = Path(temp_dir) / "logger.log"
        logger_file.touch()
        stdout_file = Path(temp_dir) / "stdout.log"
        env["DEPLOYMENT_LOG_PATH"] = str(logger_file)
        stdout_buffer = stdout_file.open("w")
        try:
            os.environ.update(env)
            if capture_output:
                sys.stdout = stdout_buffer
            ctx = get_context()
            proc = ctx.Process(
                target=SubProcess.run_ansible_playbook, args=(cwd, command)
            )
            proc.start()
            proc.join()
        finally:
            os.environ = os_env
            sys.stdout = stdout
            stdout_buffer.close()
        return SubProcess(
            proc.exitcode,
            stdout=stdout_file.read_text(),
            log=logger_file.read_text(),
        )


def run_command_with_spinner(
    cwd: str, command: List[str], env: Dict[str, str], text: str
) -> SubProcess:
    console = Console(stderr=True)
    spinner = Spinner("weather", text=text)

    with Live(spinner, refresh_per_second=3, console=console):
        try:
            result = run_command(cwd, command, env=env, capture_output=True)
        except KeyboardInterrupt:
            spinner.update(text=text + " [yellow]canceled[/yellow]")
            raise KeyboardInterrupt("User interrupted execution") from None

        if result.returncode == 0:
            spinner.update(text=text + " [green]ok[/green]")
        else:
            spinner.update(text=text + " [red]failed[/red]")
    return result


class RunnerDir(TemporaryDirectory):
    """Define and create the Ansible runner directory."""

    def __init__(self) -> None:
        super().__init__(prefix="AnsibleRunner")
        self.parent_dir = Path(self.name)
        self.env_dir = self.parent_dir / "env"
        self.inventory_dir = self.parent_dir / "inventory"
        self.project_dir = self.parent_dir / "project"
        self.aux_file_dir = self.parent_dir / "files"
        for _dir in (self.env_dir, self.inventory_dir, self.project_dir):
            _dir.mkdir(exist_ok=True, parents=True)

        self.ansible_config_file = Path(
            os.getenv("ANSIBLE_CONFIG") or self.parent_dir / "ansible.cfg"
        )
        if self.ansible_config_file.is_file():
            self.ansible_config_file.with_suffix(".cfg.tmp").write_text(
                self.ansible_config_file.read_text()
            )
        atexit.register(_del_path, self.ansible_config_file)

    def create_config(self, **kwargs: str) -> None:
        """Create an ansible config."""
        self.ansible_config_file.parent.mkdir(exist_ok=True, parents=True)
        with open(self.ansible_config_file, "w", encoding="utf-8") as stream:
            stream.write("[defaults]\n")
            for key, value in kwargs.items():
                stream.write(f"{key} = {value}\n")
            stream.write("[colors]\n")
            stream.write("included = purple\n")
            stream.write("skip = green\n")
            stream.write("[callback]\n")
            stream.write("result_format=yaml\n")
            stream.write("pretty_results=True\n")

    def create_playbook(self, content: List[Dict[str, Any]]) -> str:
        """Dump the content of a playbook into the playbook file."""
        for nn, step in enumerate(content):
            host = step.get("hosts", "")
            if not host:
                continue
            for tt, task in enumerate(step.get("tasks", [])):
                try:
                    content[nn]["tasks"][tt]["name"] = f"{host} - {task['name']}"
                except KeyError:
                    pass
        content_str = yaml.safe_dump(content)
        self.playbook_file.write_text(content_str)
        return content_str

    @property
    def ansible_exe(self) -> str:
        """Get the path to the ansible-playbook command."""
        exe = ""
        if sys.platform.lower().startswith("win"):
            exe = ".exe"
        if is_bundeled:
            return str(Path(__file__).parent / "bin" / f"ansible-playbook{exe}")
        else:
            return "ansible-playbook"

    @property
    def inventory_file(self) -> Path:
        """Define the location of the inventory file."""
        return self.project_dir / "hosts"

    @property
    def playbook_file(self) -> Path:
        """Define the location of the playbook file."""
        return self.project_dir / "playbook.yaml"

    def write_to_tempfile(self, content: str) -> str:
        """
        Write the given content to a temporary file and return the file path.
        """
        temp_file = NamedTemporaryFile(
            dir=self.parent_dir, delete=False, mode="w", suffix=".yml"
        )
        temp_file.write(content)
        temp_file.close()
        return temp_file.name

    def convert_to_file(
        self, content: Union[Path, str, List[Any], Dict[str, Any]]
    ) -> str:
        """
        Convert inventory input to a file path. If it is already a path, return it.
        If it is a string or list of dicts, write it to a temporary file and return the path.
        """
        if isinstance(content, Path):
            return str(content)
        if isinstance(content, str):
            if os.path.isfile(content):
                return content
            else:
                return self.write_to_tempfile(content)
        elif isinstance(content, (list, dict)):
            content_str = yaml.dump(json.loads(json.dumps(content)))
            return self.write_to_tempfile(content_str)
        else:
            raise ValueError("Invalid inventory format")

    def get_remote_file_content(
        self,
        host: str,
        *file_paths: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> str:
        """Get the content of a remote file.

        Parameters
        ----------
        host: str
            Remote hostname
        file_paths: str
            The paths to the file containing the target content
        username: str, default: None
            Use this username to log on
        password: str, default: None
            Instead of logging on by ssh key, use a password based log in.

        Returns
        -------
        str: Content of the file
        """
        if not host:
            return ""
        username = username or getuser()
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        file_content = ""
        try:
            logger.debug("Connecting to %s with %s", host, username)
            ssh.connect(host, username=username, password=password or None)
            for file_path in file_paths:
                _, stdout, stderr = ssh.exec_command(f"cat {file_path}")
                if stderr.channel.recv_exit_status() == 0:
                    file_content = stdout.read().decode("utf-8").strip()
                    break
        except Exception as error:
            logger.critical(
                "Couldn't establish connection to %s: %s", host, error
            )
        finally:
            ssh.close()
        return file_content

    def run_ansible_playbook(
        self,
        playbook: Optional[Union[str, Path, List[Any], Dict[str, Any]]],
        inventory: Optional[Union[str, Path, List[Any], Dict[str, Any]]],
        working_dir: Optional[Union[str, Path]] = None,
        envvars: Optional[Dict[str, str]] = None,
        extravars: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        cmdline: Optional[str] = None,
        verbosity: int = 0,
        passwords: Optional[Dict[str, str]] = None,
        hide_output: bool = False,
        text: str = "Running playbook ...",
        output: str = "",
    ) -> str:
        """
        Run an Ansible playbook using multiprocessing.

        Parameters:
        working_dir (str): Current working directory for the playbooks.
        playbook_path (str): Path to the playbook.
        inventory_path (str): Path to the inventory file.
        envvars (Optional[Dict[str, str]]): Environment variables to set for Ansible.
        extravars (Optional[Dict[str, str]]): Extra variables to pass to Ansible.
        cmdline (Optional[str]): Extra command line arguments for Ansible.
        verbosity (int): Verbosity level for Ansible output.
        tags (Optional[List[str]]): The roles that should be involved.
        hide_output (bool): Hide stdout and stderr if True, show only on failure.

        Raises:
        DeploymentError: If the Ansible playbook execution fails.
        """
        tags = tags or []
        working_dir = Path(working_dir or "").expanduser().absolute()
        playbook_path = self.convert_to_file(playbook or "")
        inventory_path = self.convert_to_file(inventory or "")
        passwords = passwords or {}
        extravars = extravars or {}
        envvars = envvars or {}
        # Prepare the command

        command = ["ansible-playbook", playbook_path, "-i", inventory_path]
        # Set environment variables
        for tag in tags:
            command += ["-t", tag]
        for key, passwd in passwords.items():
            if passwd:
                envvars[key.upper()] = passwd
                extravars[key] = f'{{{{ lookup("env", "{key.upper()}") }}}}'
        extravars["playbook_tempdir"] = str(self.aux_file_dir)
        # Add extra variables
        for key, value in extravars.items():
            command.append(f"-e {key}='{value}'")

        # Add command line arguments
        if cmdline:
            command.extend(cmdline.split())

        # Add verbosity
        if verbosity > 0:
            command.append("-" + "v" * verbosity)
        logger.debug("Running ansible command %s", " ".join(command))
        # Run the command
        if hide_output and verbosity == 0:
            result = run_command_with_spinner(
                str(working_dir), command, envvars, text
            )
        else:
            result = run_command(
                str(working_dir), command, env=envvars, capture_output=False
            )

        # Determine if the command was successful
        success = result.returncode == 0
        # Raise an error if the playbook execution failed
        if not success:
            pprint(result.stdout)
            raise DeploymentError("Deployment failed!")
        return result.log



from __future__ import annotations

from getpass import getuser
from pathlib import Path
from typing import Any, List, cast

import npyscreen

from freva_deployment import AVAILABLE_CONDA_ARCHS
from freva_deployment.utils import get_current_file_dir

from .base import (
    BaseForm,
    CheckboxInfo,
    ComboInfo,
    DictInfo,
    FileInfo,
    TextInfo,
)

DEPLOYMENT_METHODS = ["docker", "podman", "conda", "k8s"]
EXPOSE_METHODS = ["op-lb", "saas-lb", "hostport"]


def get_index(values: list[str], target: str, default: int = 0) -> int:
    """Get the target index of item in list.

    Parameters:
    ===========
    values:
        the list of values that is searched
    target:
        the item the list of values that is searched for
    default:
        if nothing is found return the default value

    Returns:
    ========
    int: Index of the the target item in the list
    """

    for n, value in enumerate(values):
        if value == target:
            return n
    return default


class CoreScreen(BaseForm):
    """Form for the core deployment configuration."""

    step: str = "core"
    """Name of this step."""
    certificates: list[str] = ["public"]
    """The type of certificate files this step needs."""

    @property
    def scheduler_systems(self):
        """Define available scheduler systems."""
        return ["local", "slurm", "pbs", "lfs", "moab", "oar", "sge"]

    def scheduler_index(self, scheduler_system: str = ""):
        """Get the index of the saved scheduler_system"""
        scheduler_system = scheduler_system or "local"
        for nn, choice in enumerate(self.scheduler_systems):
            if choice == scheduler_system:
                return nn
        return 0

    def _add_widgets(self) -> None:
        """Add widgets to the screen."""
        self.list_keys: list[str] = []
        cfg = self.get_config(self.step)
        arch = cast(str, cfg.get("arch", AVAILABLE_CONDA_ARCHS[0]))
        arch_idx = get_index(AVAILABLE_CONDA_ARCHS, arch, 0)
        self.input_fields: dict[str, tuple[npyscreen.TitleText, bool]] = dict(
            core_host=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="core",
                    key="hosts",
                    name=f"{self.num}Server Name(s) where core is deployed",
                    value=cfg.get("core_host"),
                ),
                True,
            ),
            install_dir=(
                self.add_widget_intelligent(
                    FileInfo,
                    section="core",
                    key="install_dir",
                    name=f"{self.num}Anaconda installation dir. for core",
                    value=cfg.get("install_dir", ""),
                ),
                True,
            ),
            root_dir=(
                self.add_widget_intelligent(
                    FileInfo,
                    section="core",
                    key="root_dir",
                    name=(f"{self.num}Freva configuration directory"),
                    value=cfg.get("root_dir", ""),
                ),
                False,
            ),
            base_dir_location=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="core",
                    key="base_dir_location",
                    name=(f"{self.num}User data directory"),
                    value=cfg.get("base_dir_location", ""),
                ),
                False,
            ),
            preview_path=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="core",
                    key="preview_path",
                    name=(f"{self.num}Plugin output dir for the web UI. "),
                    value=cfg.get("preview_path", ""),
                ),
                False,
            ),
            scheduler_system=(
                self.add_widget_intelligent(
                    ComboInfo,
                    section="core",
                    key="scheduler_system",
                    name=f"{self.num}Workload manager",
                    value=self.scheduler_index(
                        cast(str, cfg.get("scheduler_system"))
                    ),
                    values=self.scheduler_systems,
                ),
                True,
            ),
            scheduler_output_dir=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="core",
                    key="scheduler_output_dir",
                    name=f"{self.num}Output dir. of the scheduler system",
                    value=cfg.get("scheduler_output_dir", ""),
                ),
                False,
            ),
            admins=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="core",
                    key="admins",
                    name=f"{self.num}Set the admin user(s) - comma separated",
                    value=cfg.get("admins", getuser()),
                ),
                False,
            ),
            admin_group=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="core",
                    key="admin_group",
                    name=(
                        f"{self.num}Set the Freva admin group - "
                        "leave blank if not needed"
                    ),
                    value=cfg.get("admin_group", ""),
                ),
                False,
            ),
            ansible_become_user=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="core",
                    key="ansible_become_user",
                    name=(
                        f"{self.num}Become (sudo) user name to change to on "
                        "remote machine"
                    ),
                    value=cfg.get("ansible_become_user", ""),
                ),
                False,
            ),
            arch=(
                self.add_widget_intelligent(
                    ComboInfo,
                    section="core",
                    key="arch",
                    name=(f"{self.num}Set the target architecture"),
                    value=arch_idx,
                    values=AVAILABLE_CONDA_ARCHS,
                ),
                True,
            ),
            ansible_python_interpreter=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="core",
                    key="ansible_python_interpreter",
                    name=f"{self.num}Python path on remote machine",
                    value=cfg.get(
                        "ansible_python_interpreter", "/usr/bin/python3"
                    ),
                ),
                False,
            ),
            ansible_user=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="core",
                    key="ansible_user",
                    name=f"{self.num}Username for remote machine",
                    value=cfg.get("ansible_user", getuser()),
                ),
                False,
            ),
        )


class WebScreen(BaseForm):
    """Form for the web deployment configuration."""

    step: str = "web"
    certificates: list[str] = ["public", "private"]
    """The type of certificate files this step needs."""

    def get_index(self, choices: list[str], key: str):
        """Get the key value pair for a combo box"""
        for nn, choice in enumerate(choices):
            if choice == key:
                return nn
        return 0

    def _add_widgets(self) -> None:
        """Add widgets to the screen."""
        self.list_keys = "imprint", "scheduler_host", "allowed_hosts"
        cfg = self.get_config(self.step)

        for key in self.list_keys:
            if key in cfg and isinstance(cfg[key], str):
                value = cast(str, cfg[key])
                cfg[key] = [v.strip() for v in value.split(",") if v.strip()]
        self.input_fields: dict[str, tuple[npyscreen.TitleText, bool]] = dict(
            web_host=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="web_host",
                    name=f"{self.num}Hostname where web service is deployed on",
                    value=cfg.get("web_host"),
                ),
                True,
            ),
            data_path=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="data_path",
                    name=f"{self.num}Parent directory for any permanent data",
                    value=cast(str, cfg.get("data_path", "/opt/freva")),
                ),
                True,
            ),
            admin_user=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="admin_user",
                    name=(f"{self.num}User name that should own persistent data"),
                    value=cfg.get("admin_user", ""),
                ),
                False,
            ),
            project_website=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="project_website",
                    name=f"{self.num}Url of the Freva home page",
                    value=cfg.get("project_website", ""),
                ),
                True,
            ),
            chatbot_host=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="chatbot_host",
                    name=f"{self.num}Url to the FrevaGPT service",
                    value=cfg.get("chatbot_host", ""),
                ),
                False,
            ),
            main_color=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="main_color",
                    name=f"{self.num}Html color of the main color theme",
                    value=cfg.get("main_color", "Tomato"),
                ),
                True,
            ),
            border_color=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="border_color",
                    name=f"{self.num}Html color for the borders",
                    value=cfg.get("border_color", "#6c2e1f"),
                ),
                True,
            ),
            hover_color=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="hover_color",
                    name=f"{self.num}Html color for hover modes",
                    value=cfg.get("hover_color", "#d0513a"),
                ),
                True,
            ),
            institution_logo=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="institution_logo",
                    name=f"{self.num}Path to the institution logo",
                    value=cfg.get(
                        "institution_logo", "/path/to/logo/on/target/machine"
                    ),
                ),
                True,
            ),
            about_us_text=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="about_us_text",
                    name=f"{self.num}About us text - short blurb about Freva",
                    value=cfg.get("about_us_text", "Testing"),
                ),
                True,
            ),
            contacts=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="contacts",
                    name=f"{self.num}Contact email address",
                    value=str(cfg.get("contacts", "freva@dkrz.de")),
                ),
                True,
            ),
            imprint=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="imprint",
                    name=f"{self.num}Institution address - comma separated",
                    value=",".join(
                        cast(
                            List[str],
                            cfg.get(
                                "imprint",
                                [
                                    "freva",
                                    "German Climate Computing Centre (DKRZ)",
                                    "Bundesstr. 45a",
                                    "20146 Hamburg",
                                    "Germany",
                                ],
                            ),
                        ),
                    ),
                ),
                True,
            ),
            homepage_text=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="homepage_text",
                    name=f"{self.num}More in detail project description",
                    value=cfg.get(
                        "homepage_text",
                        (
                            "Lorem ipsum dolor sit amet, consectetur "
                            "adipiscing elit, sed do eiusmod tempor "
                            "incididunt ut labore et dolore magna aliqua. "
                            "Ut enim ad minim veniam, quis nostrud "
                            "exercitation ullamco laboris nisi ut aliquip "
                            "ex ea commodo consequat. Duis aute irure dolor "
                            "in reprehenderit in voluptate velit esse cillum "
                            "dolore eu fugiat nulla pariatur. Excepteur sint "
                            "occaecat cupidatat non proident, sunt in culpa "
                            "qui officia deserunt mollit anim id est laborum."
                        ),
                    ),
                ),
                True,
            ),
            homepage_heading=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="homepage_heading",
                    name=f"{self.num}A brief description of the project",
                    value=cfg.get(
                        "homepage_heading", "Lorem ipsum dolor sit amet"
                    ),
                ),
                True,
            ),
            scheduler_host=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="scheduler_host",
                    name=f"{self.num}Scheduler hostname(s) - comma separated",
                    value=",".join(
                        cast(
                            List[str],
                            cfg.get("scheduler_host", ["levante.dkrz.de"]),
                        )
                    ),
                ),
                True,
            ),
            allowed_hosts=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="allowed_hosts",
                    name=f"{self.num}Set additional hostnames django can serve",
                    value=",".join(
                        cast(
                            List[str],
                            cfg.get("allowed_hosts", ["localhost"]),
                        ),
                    ),
                ),
                True,
            ),
            ansible_become_user=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="ansible_become_user",
                    name=(
                        f"{self.num}Become (sudo) user name to change to on "
                        "remote machine"
                    ),
                    value=cfg.get("ansible_become_user", "root"),
                ),
                False,
            ),
            ansible_python_interpreter=(
                self.add_widget_intelligent(
                    FileInfo,
                    section="web",
                    key="ansible_python_interpreter",
                    name=f"{self.num}Pythonpath on remote machine",
                    value=cfg.get(
                        "ansible_python_interpreter", "/usr/bin/python3"
                    ),
                ),
                False,
            ),
            ansible_user=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="web",
                    key="ansible_user",
                    name=f"{self.num}Username for remote machine",
                    value=cfg.get("ansible_user", getuser()),
                ),
                False,
            ),
        )


class DBScreen(BaseForm):
    """Form for the core deployment configuration."""

    step: str = "db"

    def _add_widgets(self) -> None:
        """Add widgets to the screen."""
        self.list_keys: list[str] = []
        cfg = self.get_config(self.step)
        db_ports: list[int] = list(range(3300, 3320))
        port_idx = get_index(
            cast(List[str], list(map(str, db_ports))),
            str(cfg.get("port", 3306)),
            6,
        )
        self.input_fields: dict[str, tuple[npyscreen.TitleText, bool]] = dict(
            db_host=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="db",
                    key="db_host",
                    name=f"{self.num}Hostname where the database service is deployed",
                    value=cfg.get("db_host"),
                ),
                True,
            ),
            vault_host=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="db",
                    key="vault_host",
                    name=f"{self.num}Hostname where the vault service is deployed",
                    value=cfg.get("vault_host") or cfg.get("db_host"),
                ),
                True,
            ),
            wipe=(
                self.add_widget_intelligent(
                    CheckboxInfo,
                    section="db",
                    key="wipe",
                    max_height=2,
                    value=cfg.get("wipe", False),
                    editable=True,
                    name=(f"{self.num}Delete existing data?"),
                    scroll_exit=True,
                ),
                True,
            ),
            user=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="db",
                    key="user",
                    name=f"{self.num}Database user",
                    value=cfg.get("user", "evaluation_system"),
                ),
                True,
            ),
            db=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="db",
                    key="db",
                    name=f"{self.num}Database name",
                    value=cfg.get("db", "evaluation_system"),
                ),
                True,
            ),
            port=(
                self.add_widget_intelligent(
                    ComboInfo,
                    section="db",
                    key="port",
                    name=f"{self.num}Database Port",
                    value=port_idx,
                    values=db_ports,
                ),
                True,
            ),
            data_path=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="db",
                    key="data_path",
                    name=(f"{self.num}Parent directory for any permanent data"),
                    value=cast(str, cfg.get("data_path", "/opt/freva")),
                ),
                True,
            ),
            admin_user=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="db",
                    key="admin_user",
                    name=(f"{self.num}User name that should own persistent data"),
                    value=cfg.get("admin_user", ""),
                ),
                False,
            ),
            ansible_become_user=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="db",
                    key="ansible_become_user",
                    name=(
                        f"{self.num}Become (sudo) user name to change to on "
                        "remote machine"
                    ),
                    value=cfg.get("ansible_become_user", "root"),
                ),
                False,
            ),
            ansible_python_interpreter=(
                self.add_widget_intelligent(
                    FileInfo,
                    section="db",
                    key="ansible_python_interpreter",
                    name=f"{self.num}Pythonpath on remote machine",
                    value=cfg.get(
                        "ansible_python_interpreter", "/usr/bin/python3"
                    ),
                ),
                False,
            ),
            ansible_user=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="db",
                    key="ansible_user",
                    name=f"{self.num}Username for remote machine",
                    value=cfg.get("ansible_user", getuser()),
                ),
                False,
            ),
        )


class FrevaRestScreen(BaseForm):
    """Form for the freva-rest deployment configuration."""

    step: str = "freva_rest"

    def _add_widgets(self) -> None:
        """Add widgets to the screen."""
        self.list_keys: list[str] = []
        cfg = self.get_config(self.step)
        freva_rest_ports: list[int] = list(range(7770, 7780))
        solr_mem_values = [f"{i}g" for i in range(1, 10)]
        solr_mem_select = get_index(
            solr_mem_values, cast(str, cfg.get("solr_mem", "4g")), 3
        )
        freva_rest_port_idx = get_index(
            [str(p) for p in freva_rest_ports],
            str(cfg.get("freva_rest_port", 7777)),
            7,
        )
        self.input_fields: dict[str, tuple[npyscreen.TitleText, bool]] = dict(
            freva_rest_host=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="freva_rest",
                    key="freva_rest_host",
                    name=(
                        f"{self.num}Server Name(s) where the "
                        "freva-rest service is deployed"
                    ),
                    value=cfg.get("freva_rest_host", ""),
                ),
                True,
            ),
            search_server_host=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="freva_rest",
                    key="redis_host",
                    name=f"{self.num} Set the index server host name",
                    value=cast(
                        str,
                        cfg.get("search_server_host")
                        or cfg.get("freva_rest_host"),
                    ),
                ),
                False,
            ),
            mongodb_server_host=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="freva_rest",
                    key="redis_host",
                    name=f"{self.num} Set the mongoDB host name",
                    value=cast(
                        str,
                        cfg.get("mongodb_server_host")
                        or cfg.get("freva_rest_host"),
                    ),
                ),
                False,
            ),
            redis_host=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="freva_rest",
                    key="redis_host",
                    name=f"{self.num} Set the redis host name",
                    value=cast(
                        str,
                        cfg.get("redis_host") or cfg.get("freva_rest_host"),
                    ),
                ),
                False,
            ),
            wipe=(
                self.add_widget_intelligent(
                    CheckboxInfo,
                    section="freva_rest",
                    key="wipe",
                    max_height=2,
                    value=cfg.get("wipe", False),
                    editable=True,
                    name=(f"{self.num}Delete existing data?"),
                    scroll_exit=True,
                ),
                True,
            ),
            solr_mem=(
                self.add_widget_intelligent(
                    ComboInfo,
                    section="freva_rest",
                    key="solr_mem",
                    name=(
                        f"{self.num}Virtual memory (in GB) for the search "
                        "engine service"
                    ),
                    value=solr_mem_select,
                    values=solr_mem_values,
                ),
                True,
            ),
            freva_rest_port=(
                self.add_widget_intelligent(
                    ComboInfo,
                    section="freva_rest",
                    key="freva_rest_port",
                    name=f"{self.num}Freva-rest API port",
                    value=freva_rest_port_idx,
                    values=freva_rest_ports,
                ),
                True,
            ),
            admin_user=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="freva_rest",
                    key="admin_user",
                    name=(f"{self.num}User name that should own persistent data"),
                    value=cfg.get("admin_user", ""),
                ),
                False,
            ),
            data_path=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="freva_rest",
                    key="data_path",
                    name=f"{self.num}Parent directory for any permanent data",
                    value=cast(str, cfg.get("data_path", "/opt/freva")),
                ),
                True,
            ),
            data_loader_portal_hosts=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="freva_rest",
                    key="data_loader_portal_hosts",
                    name=(
                        f"{self.num} ',' separated hostname(s) data-loading portal "
                        "to provide zarr. Leave blank for none"
                    ),
                    value=cast(str, cfg.get("data_loader_portal_hosts", "")),
                ),
                False,
            ),
            oidc_url=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="freva_rest",
                    key="oidc_url",
                    name=(f"{self.num}Config url of the OIDC service"),
                    value=cast(str, cfg.get("oidc_url", "")),
                ),
                True,
            ),
            oidc_client=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="freva_rest",
                    key="oidc_client",
                    name=(f"{self.num}Name of the OIDC client (app name)"),
                    value=cast(str, cfg.get("oidc_client", "freva")),
                ),
                True,
            ),
            oidc_client_secret=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="freva_rest",
                    key="oidc_client_secret",
                    name=(f"{self.num}OIDC client secret. Leave blank for none"),
                    value=cast(str, cfg.get("oidc_client_secret", "")),
                ),
                False,
            ),
            oidc_token_claims=(
                self.add_widget_intelligent(
                    DictInfo,
                    section="freva_rest",
                    key="oidc_token_claims",
                    name=f"{self.num}OIDC authorization filters",
                    value=cast(
                        dict[str, list[str]], cfg.get("oidc_token_claims", {})
                    ),
                ),
                False,
            ),
            ansible_become_user=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="freva_rest",
                    key="ansible_become_user",
                    name=(
                        f"{self.num}Become (sudo) user name to change to on "
                        "remote machine"
                    ),
                    value=cfg.get("ansible_become_user", "root"),
                ),
                False,
            ),
            ansible_python_interpreter=(
                self.add_widget_intelligent(
                    FileInfo,
                    section="freva_rest",
                    key="ansible_python_interpreter",
                    name=f"{self.num}Pythonpath on remote machine",
                    value=cfg.get(
                        "ansible_python_interpreter", "/usr/bin/python3"
                    ),
                ),
                False,
            ),
            ansible_user=(
                self.add_widget_intelligent(
                    TextInfo,
                    section="freva_rest",
                    key="ansible_user",
                    name=f"{self.num}Username for remote machine",
                    value=cfg.get("ansible_user", getuser()),
                ),
                False,
            ),
        )


class RunForm(npyscreen.FormMultiPageAction):
    """Definition of the form that applies the actual deployment"""

    _num: int = 0

    @property
    def num(self) -> str:
        """Calculate the number for enumerations of any input field."""
        self._num += 1
        return f"{self._num}. "

    def on_ok(self) -> None:
        """Define what happens once the `ok` for applying the deployment is hit."""

        if not self.project_name.value:
            npyscreen.notify_confirm(
                "You have to set a project name", title="ERROR"
            )
            return
        missing_form: None | str = self.parentApp.check_missing_config()
        if missing_form:
            self.parentApp.change_form(missing_form)
            return
        public_keyfile = self.public_keyfile.value
        cert_files = dict(
            public=public_keyfile or "",
            private=self.private_keyfile.value or "",
        )
        save_file = Path(
            self.parentApp.get_save_file(self.inventory_file.value or None)
        )
        if isinstance(self.gen_keys.value, list):
            gen_keys = bool(self.gen_keys.value[0])
        else:
            gen_keys = bool(self.gen_keys.value)
        for key_type, keyfile in cert_files.items():
            key_file = Path(get_current_file_dir(save_file.parent, str(keyfile)))
            for step, deploy_form in self.parentApp._forms.items():
                if not keyfile or not Path(key_file).is_file():
                    if (
                        key_type in deploy_form.certificates
                        and step in self.parentApp.steps
                        and gen_keys is False
                    ):
                        if keyfile:
                            msg = f"{key_type} certificate file `{key_file}` must exist."
                        else:
                            msg = f"You must give a {key_type} certificate file"
                        npyscreen.notify_confirm(msg, title="ERROR")
                        return
        _ = self.parentApp.save_config_to_file(
            save_file=save_file, write_toml_file=True
        )
        try:
            ssh_port = int(self.ssh_port.value)
        except ValueError:
            ssh_port = 22
        self.parentApp.setup = {
            "steps": list(set(self.parentApp.steps)),
            "ask_pass": bool(self.use_ssh_pw.value),
            "config_file": str(save_file) or None,
            "ssh_port": ssh_port,
            "skip_version_check": bool(self.skip_version_check.value),
            "local_debug": bool(self.local_debug.value),
            "gen_keys": bool(gen_keys),
        }
        self.parentApp.thread_stop.set()
        self.parentApp.exit_application(
            save_file=save_file, msg="Do you want to continue?"
        )

    def on_cancel(self) -> None:
        """Define what happens after the the cancel button is hit."""
        name = self.parentApp.current_form.lower()
        self.parentApp.setup = {}
        for step, form_name in self.parentApp._steps_lookup.items():
            if name.startswith(step):
                # Tell the MyTestApp object to change forms.
                self.parentApp.change_form(form_name)
                return
        self.parentApp.change_form("MAIN")

    def create(self) -> None:
        """Custom definitions executed when the from gets created."""
        self.how_exited_handers[npyscreen.wgwidget.EXITED_ESCAPE] = (
            self.parentApp.exit_application
        )
        self.add_handlers({"^F": self.show_info})
        self._add_widgets()

    def show_info(self, *args: Any, **kwargs: Any) -> None:
        """Display an info if present."""
        if isinstance(self.current_info, str):
            npyscreen.notify_confirm(
                self.current_info, title="Detailed Information"
            )

    def _add_widgets(self) -> None:
        """Add the widgets to the form."""

        project_name = self.parentApp.config.get(
            "project_name", self.parentApp._read_cache("project_name", "")
        )
        self.project_name = self.add_widget_intelligent(
            npyscreen.TitleText,
            name=f"{self.num}Set the name of the project",
            value=project_name,
        )
        self.deployment_method = self.add_widget_intelligent(
            ComboInfo,
            key="deployment_method",
            name=f"{self.num}Deployment Method",
            info=(
                "The `deployment_method` key sets the option of how the "
                "installation of the service is realised. `docker`, "
                "`podman` leverages podman or docker, `conda` uses "
                "conda-forge to install the service while `k8s` "
                "involves a kubernetes based deployment "
                'Chosse between: "docker", "podman", "conda", "k8s"'
            ),
            value=get_index(
                DEPLOYMENT_METHODS,
                cast(
                    str,
                    self.parentApp.config.get("deployment_method", "docker"),
                ),
            ),
            values=DEPLOYMENT_METHODS,
        )
        self.inventory_file = self.add_widget_intelligent(
            npyscreen.TitleFilename,
            name=f"{self.num}Save config as",
            value=str(self.parentApp.save_file or ""),
        )
        self.public_keyfile = self.add_widget_intelligent(
            npyscreen.TitleFilename,
            name=f"{self.num}Select a public certificate file; needed for steps web, core, db",
            value=self.parentApp.read_cert_file("public_keyfile"),
        )
        self.private_keyfile = self.add_widget_intelligent(
            npyscreen.TitleFilename,
            name=f"{self.num}Select a private certificate file; needed for steps web",
            value=self.parentApp.read_cert_file("private_keyfile"),
        )
        self.gen_keys = self.add_widget_intelligent(
            npyscreen.RoundCheckBox,
            max_height=2,
            value=self.parentApp._read_cache("gen_keys", False),
            name=f"{self.num}Generate a pair web certificates, debugging",
        )
        self.k8s_deploy_host = self.add_widget_intelligent(
            TextInfo,
            key="deploy_host",
            section="kubernetes",
            name=f"{self.num}Set a deploy host",
            value=self.parentApp.config.get("kubernetes", {}).get("deploy_host"),
        )

        self.k8s_expose_method = self.add_widget_intelligent(
            ComboInfo,
            key="expose_method",
            section="kubernetes",
            name=f"{self.num}K8s Expose Method",
            value=get_index(
                EXPOSE_METHODS,
                cast(
                    str,
                    self.parentApp.config.get("kubernetes", {}).get(
                        "expose_method",
                        "op-lb",
                    ),
                ),
            ),
            values=EXPOSE_METHODS,
        )
        self.use_ssh_pw = self.add_widget_intelligent(
            npyscreen.RoundCheckBox,
            max_height=2,
            editable=True,
            value=self.parentApp._read_cache("ssh_pw", False),
            name=f"{self.num}Use password for ssh connection",
            scroll_exit=True,
        )
        self.skip_version_check = self.add_widget_intelligent(
            npyscreen.RoundCheckBox,
            max_height=2,
            editable=True,
            value=self.parentApp._read_cache("skip_version_check", False),
            name=f"{self.num}Skip the version check, use with caution",
        )
        self.ssh_port = self.add_widget_intelligent(
            npyscreen.TitleText,
            name=f"{self.num}If you need to, you can change the ssh port here",
            value=str(self.parentApp._read_cache("ssh_port", 22)),
        )
        self.local_debug = self.add_widget_intelligent(
            npyscreen.RoundCheckBox,
            max_height=2,
            value=self.parentApp._read_cache("local_debug", False),
            editable=True,
            name=f"{self.num}Deploy services on the local machine, debug",
        )



"""

This resAPI let's you safely store and retrieve secrets. Such secrets include
the login credentials of the freva database or user name and password for
the django enmail interface.

Overview
--------

"""

import argparse
import base64
import json
import logging
import os
import pathlib
import random
import sys
import time
from subprocess import Popen
from typing import Annotated, Dict, List, Optional, TypedDict, cast

import hvac
import requests
from fastapi import Body, FastAPI, Header, HTTPException, Path, status
from fastapi.responses import JSONResponse

KeyType = TypedDict("KeyType", {"keys": List[str], "token": str})
if os.getenv("KEY_FILE"):
    KEY_FILE = pathlib.Path(os.getenv("KEY_FILE"))
elif os.getenv("CONDA_PREFIX"):
    KEY_FILE = (
        pathlib.Path(os.getenv("CONDA_PREFIX"))
        / "share"
        / "freva-rest-server"
        / "vault"
        / "keys"
    )
else:
    KEY_FILE = pathlib.Path("/vault/file/keys")
VAULT_ADDR = os.environ.get("VAULT_ADDR", "http://127.0.0.1:8200")
POLICY = """
    path "secret" {
        capabilities = ["create", "read", "update", "delete", "list"]
    }
"""

PHRASES = [
    "There seems to be a noose around this request.",
    "Our guns jammed! Something went wrong.",
    "When you have to shoot, shoot, don't talk.",
    "The bullets didn't fire.",
    "I've never seen so many men wasted so badly.",
    "You're not digging.",
    "We're in the desert without a horse.",
    "We faced a showdown, and it didn't end well."
    "Blondie! You know what you are? Just a dirty son of a...",
]

metadata_tags = [
    {
        "name": "Secrets",
        "description": "Read and Store secret key=value pairs.",
    }
]

VERSION = "2505.0.3"

app = FastAPI(
    title="freva-vault",
    debug=False,
    summary="The freva secret restAPI 🔒",
    description=__doc__,
    openapi_url="/vault/docs/openapi.json",
    docs_url="/vault/docs",
    version=VERSION,
    redoc_url=None,
    contact={"name": "DKRZ, Clint", "email": "freva@dkrz.de"},
    license_info={
        "name": "BSD 2-Clause License",
        "url": "https://opensource.org/license/bsd-2-clause",
    },
)
logging.basicConfig(
    format="%(asctime)s - %(name)s - [%(levelname)s] - %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%dT%H:%M:%S%z",
)
logger = logging.getLogger("secret-reader")


def cli() -> str:
    """Set up the command line interface."""
    parser = argparse.ArgumentParser(
        prog=sys.argv[0],
        description="Start the vault server.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-c",
        "--config",
        type=pathlib.Path,
        help="Vault config path.",
        default="/opt/vault/vault-server-tls.hcl",
    )
    parser.add_argument(
        "-v", "--verbose", action="count", help="Verbosity level", default=0
    )
    args = parser.parse_args()
    logger.setLevel(max(logging.INFO - 10 * args.verbose, logging.DEBUG))
    return str(args.config.expanduser().absolute())


class VaultClient:
    """Interact with the client."""

    secret_threshold: int = 5
    """The number of shares required to reconstruct the master key."""
    secret_shares: int = 5
    """The number of shares to split the master key into."""
    client = hvac.Client(
        url=VAULT_ADDR,
    )
    keys: KeyType = {"token": "", "keys": []}

    def __init__(self) -> None:
        prefix = "VAULT_ADD_VAR_"
        self.add_keys = {}
        for key, value in os.environ.items():
            if key.startswith(prefix):
                k = key.removeprefix(prefix).lower().replace("_", ".")
                self.add_keys[k] = value

    def _auth_vault(self) -> None:
        if not self.client.is_authenticated():
            keys = self.unseal()
            self.client.token = keys.get("token")

    def update_secret(self, path: str, **secret: str) -> None:
        """Update or create a secret."""
        self._auth_vault()
        old_secret = self.get_secret(path) or {}
        old_secret.update(secret)
        self.client.secrets.kv.v1.create_or_update_secret(
            path=path, secret=old_secret
        )

    def get_secret(self, path: str) -> Optional[Dict[str, str]]:
        """Get the secretes of a path."""
        self._auth_vault()
        try:
            return cast(
                Optional[Dict[str, str]],
                self.client.secrets.kv.v1.read_secret(path).get("data"),
            )
        except hvac.exceptions.VaultError:
            logger.warning("Could not find secret path: %s", path)
            return None

    @property
    def token(self) -> str:
        """Get the root token."""
        self._auth_vault()
        return self.client.token or ""

    @classmethod
    def init_vault(cls) -> KeyType:
        """Setup a fresh vault."""
        if cls.client.sys.is_initialized() is False:
            keys = cls.client.sys.initialize(
                cls.secret_shares, cls.secret_threshold
            )
            keys.pop("keys_base64", "")
            keys["token"] = keys.pop("root_token")
            KEY_FILE.parent.mkdir(exist_ok=True, parents=True)
            KEY_FILE.write_bytes(
                base64.b64encode(json.dumps(keys).encode("utf-8"))
            )
        elif not KEY_FILE.is_file():
            logger.critical(
                "Vault is initialized but the key file does not exist"
            )
            return {"token": "", "keys": []}
        return cast(
            KeyType,
            json.loads(base64.b64decode(KEY_FILE.read_bytes())),
        )

    @property
    def vault_state(self) -> str:
        """Get a human readable state of the vault."""
        try:
            is_sealed = Vault.client.sys.is_sealed()
        except requests.exceptions.ConnectionError:
            is_sealed = None
        return {True: "sealed", False: "unsealed", None: "down"}[is_sealed]

    @classmethod
    def configure(cls, token: str) -> None:
        """Configure the vault if not already done."""
        cls.client.token = token
        if "secret/" not in cls.client.sys.list_mounted_secrets_engines():
            cls.client.sys.enable_secrets_engine("kv", "secret")
            cls.client.sys.create_or_update_policy("secret", POLICY)

    @classmethod
    def unseal(cls) -> KeyType:
        """Unseal the vault."""

        logger.info("Unsealing vault")
        for _ in range(5):
            try:
                keys = cls.init_vault()
                if cls.client.sys.is_sealed() is True:
                    cls.client.sys.submit_unseal_keys(keys.get("keys", []))
                if cls.client.sys.is_sealed() is False:
                    logger.info("Vault has been unsealed.")
                    cls.configure(keys["token"])
                    return keys
            except requests.exceptions.ConnectionError:
                logger.warning("Vault not ready yet.")
                time.sleep(1)
            except Exception as error:
                logger.error("Failed to unseal vault: %s", error)
                return {"token": "", "keys": []}
        logger.error("Vault doesn't seem to be up.")
        return {"token": "", "keys": []}


Vault = VaultClient()


@app.get("/vault/status", tags=["Secrets"])
async def get_vault_status() -> JSONResponse:
    """Get the status of the vault."""
    return JSONResponse(
        content={"status": Vault.vault_state, "version": VERSION}, status_code=200
    )


@app.post("/vault/{path}", tags=["Secrets"])
async def update_secret(
    path: Annotated[str, Path(description="Secret location.", examples=["test"])],
    secrets: Annotated[
        Optional[Dict[str, str]],
        Body(
            description=(
                "The secret that should be stored, "
                "this is a string representation of the"
                " the secrets. Secrets are encoded via "
                "key=value. Multiple secrets are ',' "
                "comma separated."
            ),
            examples=[{"foo": "bar", "hoo": "roo"}],
        ),
    ] = None,
    admin_pw: Annotated[
        Optional[str],
        Header(
            alias="password",
            description="Give the pre defined admin password.",
            examples="password",
            tile="Password",
        ),
    ] = None,
) -> JSONResponse:
    """Update or create a secret."""
    if path == "test":
        return JSONResponse(
            {"message": "success"}, status_code=status.HTTP_201_CREATED
        )
    if admin_pw != os.environ.get("ROOT_PW", ""):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Permission denied",
        ) from None
    if not secrets:
        JSONResponse(
            content={"message": random.choice(PHRASES)},
            status_code=status.HTTP_204_NO_CONTENT,
        )
    try:
        Vault.update_secret(path, **secrets)
    except hvac.exceptions.VaultError:
        logger.warning("Could not add secrets %s to %s", path, secrets)
        raise HTTPException(
            detail=random.choice(PHRASES),
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        ) from None
    return JSONResponse(
        content={"message": "success"}, status_code=status.HTTP_201_CREATED
    )


@app.get("/vault/{path}/{public_key}", tags=["Secrets"])
async def read_secret(
    path: Annotated[
        str,
        Path(
            description="The name of the k/v secrets path",
            examples="data",
        ),
    ],
    public_key: Annotated[
        str,
        Path(
            description="hexdigest representation of the sha512 freva public key.",
            examples="foo",
        ),
    ],
) -> JSONResponse:
    """Read evaluation system secrets from the vault."""
    status_code = 400
    if len(public_key) != 128:  # This is not a checksum of a cert.
        text = f"But the vault is {Vault.vault_state}"
        raise HTTPException(
            detail=f"{random.choice(PHRASES)} {text}", status_code=status_code
        ) from None
    # Get the information from the vault
    data = Vault.get_secret(path)
    if data is not None:
        status_code = 200
    return JSONResponse(content=data or {}, status_code=status_code)


if __name__ == "__main__":

    Popen(["vault", "server", "-config", cli()])
    key = Vault.unseal()
    if key.get("token") and Vault.add_keys:
        print(Vault.add_keys)
        Vault.update_secret("data", **Vault.add_keys)



#!/usr/bin/env python3
"""Simple script to download data via python build in modules."""

import argparse
import logging
import urllib.request
from pathlib import Path
from typing import Tuple

logging.basicConfig(
    format="%(name)s: %(message)s",
    level=logging.INFO,
)


logger = logging.getLogger("downloader")


def reporthook(count: float, block_size: float, total_size: float) -> None:
    """Print the download status."""
    if count == 0:
        return
    frac = count * block_size / total_size
    percent = int(100 * frac)
    bar = "#" * int(frac * 40)
    msg = "Downloading: [{0:<{1}}] | {2}% Completed".format(
        bar, 40, round(percent, 2)
    )
    print(msg, end="\r", flush=True)
    if frac >= 1:
        print()


def cli() -> Tuple[str, str]:
    """Parse the command line arguments."""

    app = argparse.ArgumentParser(
        description="Download files.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    app.add_argument(
        "url",
        help="The url of the file that needs to be downloaded.",
        type=str,
    )
    app.add_argument(
        "-o",
        "--output",
        help="The output file name.",
        type=str,
    )
    app.add_argument(
        "-v",
        "--verbose",
        help="Add debug output",
        action="store_true",
        default=False,
    )
    args = app.parse_args()
    if args.verbose:
        logger.setLevel(logging.DEBUG)
    return args.url, args.output or "output"


def download(url: str, output: str) -> None:
    """Download a file."""
    out = Path(output).expanduser()
    out.parent.mkdir(exist_ok=True, parents=True)
    logger.debug("Downloading %s to %s", url, out)
    urllib.request.urlretrieve(
        url,
        filename=str(out),
        reporthook=reporthook,
    )
    out.chmod(0o755)


if __name__ == "__main__":
    download(*cli())



import locale
import os
import re
import sys
from pathlib import Path

if sys.platform.lower().startswith("win"):
    getlocale = locale.getlocale
    getfilesystemencoding = sys.getfilesystemencoding
    locale.getlocale = lambda: ("utf-8", "utf-8")
    sys.getfilesystemencoding = lambda: "utf-8"
    os.get_blocking = lambda x: True
    try:
        import ansible

        ansible_cli_path = Path(ansible.__file__).parent / "cli" / "__init__.py"
    finally:
        sys.getfilesystemencoding = getfilesystemencoding
        locale.getlocale = getlocale
    ansible_cli_path.write_text(
        re.sub("raise SystemExit", "print", ansible_cli_path.read_text())
    )
    for inp_file in Path(ansible.__file__).parent.rglob("*.py"):
        content = inp_file.read_text(encoding="utf-8")
        write = False
        for call in ("isatty", "get_blocking"):
            if f"os.{call}" in content:
                write = True
                if "import os, termios" not in content:
                    content = content.replace("import os", "import os, termios")
                content = content.replace(f"os.{call}", f"termios.{call}")
        if "get_context('fork')" in content or 'get_context("fork")' in content:
            write = True
            content = content.replace("get_context('fork')", "get_context('spawn')")
            content = content.replace('get_context("fork")', 'get_context("spawn")')
        if "RE_TASKS =" in content:
            for line in content.splitlines():
                if line.startswith("RE_TASKS"):
                    break
            content = content.replace(line, "RE_TASKS = re.compile(u'(?:^|)+tasks?$')")
            write = True
        if write:
            inp_file.write_text(content, encoding="utf-8")
    display = Path(ansible.__file__).parent / "utils" / "display.py"
    content_l = []
    for line in display.read_text(encoding="utf-8").splitlines():
        if line.startswith("_LIBC ="):
            line = ""
        if "fcntl.ioctl" in line:
            line = line.replace(line.strip(), f"tty_size = fcntl.get_terminal_size()")
        content_l.append(line)

    content = (
        "\n".join(content_l)
        .replace("import ctypes.util", "import ctypes.util\nimport wcwidth")
        .replace("_LIBC", "wcwidth")
    )
    display.write_text(content, encoding="utf-8")
else:
    import PyInstaller.depend.bindepend

    bindepend = Path(PyInstaller.depend.bindepend.__file__)
    bindepend.write_text(
        re.sub(
            "encoding='utf-8',",
            "encoding='utf-8', errors='ignore',",
            bindepend.read_text(),
        )
    )



"""Definition of custom Errors."""

import sys
from functools import partial, wraps
from types import TracebackType
from typing import Any, Callable, Optional, Type

from rich import print as pprint

from .logger import logger


class DeploymentError(BaseException):
    """Define an error for when the deployment goes wrong.

    For this error suppresses the stack trace, since it has nothing to do
    with the code."""

    def __init__(self, message: str = "") -> None:
        print(message)
        logger.error(f"{self.__class__.__name__}: Deployment failed!")
        self.error = message

    def __str__(self) -> str:
        return ""

    def __rep__(self) -> str:
        return f"{self.__class__.__name__}: {self.error}"


class ConfigurationError(BaseException):
    """Define an error when something was mis-configured."""

    def __init__(self, message: str = "") -> None:
        self.error = message

    def __str__(self) -> str:
        return ""

    def __rep__(self) -> str:
        return f"{self.__class__.__name__}: {self.error}"


def handled_exception(func: Callable[..., Any]) -> Callable[..., Any]:
    """Wrap the exception handler around a function."""

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        """Wrapper function that handles the exception."""
        sys.excepthook = partial(exception_hook, _sys_handler=sys.excepthook)
        try:
            return func(*args, **kwargs)
        except (ConfigurationError, DeploymentError) as exception:
            raise strip_traceback(exception) from None

    return wrapper


def exception_hook(
    exception_type: Type[BaseException],
    exception: BaseException,
    traceback: TracebackType,
    _sys_handler: Callable[
        [Type[BaseException], BaseException, Optional[TracebackType]], Any
    ] = sys.excepthook,
) -> Any:
    """Make certain types of exceptions silent."""
    if isinstance(exception, ConfigurationError):
        while traceback.tb_next:
            traceback = traceback.tb_next
        code = traceback.tb_frame.f_code
        msg = (
            "Traceback (offending line only):\n"
            f'File "{code.co_filename}", in line {traceback.tb_lineno}, '
            f"in {code.co_name}"
        )
        logger.critical("%s\n%s", exception.__rep__(), msg)
        return exception
    if isinstance(exception, DeploymentError):
        return exception
    return _sys_handler(exception_type, exception, traceback)


def strip_traceback(exception: BaseException) -> BaseException:
    """Get rid of all the tracbacks, expect the last one."""
    traceback = exception.__traceback__
    if traceback is None:
        return exception
    while traceback.tb_next:
        traceback = traceback.tb_next
    exception.__traceback__ = traceback
    return exception



"""Define the version of the microservices."""

import argparse
import json
import os
import ssl
import sys
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List
from urllib.request import urlopen

from appdirs import user_cache_dir
from packaging.version import Version
from rich import print as pprint
from rich.prompt import Prompt

from .error import ConfigurationError, handled_exception

ssl_context = ssl._create_unverified_context()


class VersionAction(argparse._VersionAction):
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string=None,
    ):
        version = self.version or "%(prog)s"
        pprint(version % {"prog": parser.prog or sys.argv[1]})
        parser.exit()


def display_versions() -> str:
    """Get all service versions for display."""
    minimum_version = get_versions()
    lookup = {
        "freva_rest": "freva-rest API",
        "core": "freva-core",
        "solr": "Apache Solr",
        "web": "webUI",
        "vault": "Freva Vault",
        "mongodb_server": "MongoDB",
        "db": "MySQL",
    }
    versions = ""
    for service, version in minimum_version.items():
        service_name = lookup.get(service, service)
        versions += f"\n   [b][green]{service_name}[/b][/green] {version}"
    return versions


def _download(url: str) -> str:
    try:
        with urlopen(url, context=ssl_context) as res:
            return res.read().decode()
    except Exception as error:
        raise ConfigurationError(f"Could not download {url}: {error}")


@handled_exception
def get_versions(_versions: List[Dict[str, str]] = []) -> Dict[str, str]:
    """Read the necessary versions of microservices."""
    if _versions:
        return _versions[0]

    _versions.append(
        json.loads((Path(__file__).parent / "versions.json").read_text())
    )
    url = (
        "https://raw.githubusercontent.com/freva-org/freva-service-config"
        "/refs/heads/main/{service}/requirements.txt"
    )

    for service in ("mongo", "solr", "nginx", "redis"):
        for line in _download(url.format(service=service)).splitlines():
            if not line.startswith("#") and "=" in line:
                version = line.strip().split("=")[-1]
                _versions[0][service] = version
                break
    _versions[0]["mongodb_server"] = _versions[0].pop("mongo")
    return _versions[0]


def get_steps_from_versions(detected_versions: Dict[str, str]) -> List[str]:
    """Decide on services the should be deployed, based on their versions.

    Parameters
    ----------
    detected_versions: dict
        The versions that have been detected to be deployed.

    Returns
    -------
    list: A list of services that should be updated.
    """
    minimum_version = get_versions()
    steps = []
    for service in ("web", "vault", "freva_rest", "core"):
        if service == "vault":
            service_name = "db"
        else:
            service_name = service
        min_version = Version(minimum_version[service].strip("v"))
        version_str = detected_versions.get(service, "").strip("v").strip()
        if service == "web" and not version_str:
            continue
        version = Version(version_str or "0.0.0")
        if version < min_version:
            steps.append(service_name)
        elif version > min_version:
            # We do have a problem: an installed version has a higher version
            # the the defined minimum version, possibly the deployment
            # software is outdated.
            if os.environ.get("INTERACTIVE_DEPLOY", "1"):
                answ = (
                    Prompt.ask(
                        f"The installed version for [green]{service}[/green] is higher"
                        " than the min. defined version.\nThere might be"
                        " a chance that the current deployment software"
                        " is outdated.\nIf you continue you will "
                        f"[b]downgrade[/b] [green]{service}[/green] from "
                        f"{version} to {min_version}. "
                        "\nDo you want to continue \\[y|N]"
                    ).lower()
                    or "n"
                )
            else:
                answ = "n"
            if answ[0] == "y":
                steps.append(service_name)
            else:
                raise SystemExit(1)
    return steps



"""Command line interface for creating compose files."""

from __future__ import annotations

import argparse
import base64
import re
from base64 import b64encode
from pathlib import Path
from typing import Optional

import namegenerator
import yaml
from rich_argparse import ArgumentDefaultsRichHelpFormatter

from freva_deployment import __version__

from ..deploy import DeployFactory
from ..logger import logger, set_log_level
from ..utils import RichConsole, asset_dir, config_dir

COMPOSE_TASK = """---
- name: Render compose file locally only
  hosts: all
  connection: local
  gather_facts: no

  tasks:
    - name: Template out docker-compose.yml
      template:
        src: {asset_dir}/playbooks/templates/service-compose.yml.j2
        dest: {pwd}/{project_name}-compose.yml
"""

SYSTEMD_TMPL = """
[Unit]
Description=Start/Stop freva services containers
After=network-online.target
Wants=network-online.target
[Service]
TimeoutStartSec=35s
TimeoutStopSec=35s
ExecStartPre=/usr/bin/env sh -c "{engine} compose --project-name {project_name} -f <compose-dir>/{project_name}-compose.yml down --remove-orphans"
ExecStart=/usr/bin/env sh -c "{engine} compose --project-name {project_name} -f <compose-dir>/{project_name}-compose.yml up --remove-orphans"
ExecStop=/usr/bin/env sh -c "{engine} compose --project-name {project_name} -f <compose-dir>/{project_name}-compose.yml down --remove-orphans"
Restart=on-failure
RestartSec=5
StartLimitBurst=5
[Install]
WantedBy=default.target

"""


def comment_entries(toml_str, entries_to_comment):
    lines = toml_str.splitlines()
    result = []
    for line in lines:
        stripped = line.lstrip()
        # If line starts with one of the target entries, comment it
        for entry in entries_to_comment:
            if re.match(rf'^\["{re.escape(entry)}",', stripped):
                indent = line[: len(line) - len(stripped)]
                result.append(f"{indent}#    {stripped}")
                break
        else:
            result.append(line)
    return "\n".join(result)


def create_compose(args: argparse.Namespace) -> None:
    """Create a compose file."""
    set_log_level(args.verbose)
    with DeployFactory(
        steps=None,
        config_file=args.config_file,
        local_debug=False,
        gen_keys=True,
    ) as DF:

        eval_conf_enc = b64encode(
            DF.create_eval_config().read_text().encode()
        ).decode()
        extra = {
            "eval_config_content": eval_conf_enc,
            "use_core": args.no_plugins is False,
            "uid": args.user,
            "redis_password": DF._create_random_passwd(30, 10),
            "redis_username": namegenerator.gen(),
        }
        logger.info("Parsing configurations")
        inventory = yaml.safe_load(DF.parse_config(DF.steps, **extra))
        if args.no_plugins is True:
            web_config = base64.b64decode(
                inventory["web"]["vars"]["web_config_content"]
            ).decode()
            to_comment = ["Plugins", "History", "Result-Browser"]
            web_config = comment_entries(web_config, to_comment)
            inventory["web"]["vars"]["web_config_content"] = base64.b64encode(
                web_config.encode()
            ).decode()
            plugin_note = ""
        else:
            plugin_note = (
                "[b red]:bulb: NOTE:[/] If want to use plugins you"
                " should install the freva libraries via pip "
                "or conda:\n\n"
                "  [b]python -m pip install freva-client freva[/b] (or) \n"
                "  [b]conda -c conda-forge install freva-client freva [/b]\n\n"
                "You should then set the "
                "[b]EVALUATIO_SYSTEM_CONFIG_FILE[/b] env variable "
                "for the [b]web-server[/b] section in the compose file to the "
                "config file that was installed by conda/pip - e.g\n\n  "
                "<base-path-to-python-env>/freva/"
                "evaluation_system.conf\n"
                "This path also needs to be mounted as a volume into the "
                "container.\n"
            )

        playbook = COMPOSE_TASK.format(
            pwd=Path.cwd(), project_name=DF.project_name, asset_dir=asset_dir
        )
        web_conf = (
            Path(inventory["core"]["vars"]["core_root_dir"])
            / "share"
            / "freva"
            / "web"
            / "freva_web.toml"
        )
        for key in inventory:
            inventory[key]["hosts"] = "localhost"
        inventory["web"]["vars"]["web_config_file"] = str(web_conf)
        logger.debug(yaml.safe_dump(inventory))

        DF._td.run_ansible_playbook(
            working_dir=asset_dir / "playbooks",
            playbook=playbook,
            inventory=inventory,
            verbosity=args.verbose,
        )
        yml_file = Path.cwd() / f"{DF.project_name}-compose.yml"
        service_file = yml_file.with_suffix(".service")
        config_path = (
            Path(inventory["core"]["vars"]["core_root_dir"])
            / "freva"
            / "web"
            / "freva_web.toml"
        )

        RichConsole.rule("")
        RichConsole.print(
            (
                f"The compose file ({yml_file.name}) has been created. "
                "You can copy the file to your server and start "
                f"the compose command.\n\n{plugin_note}"
                "The web config file will be located in the "
                f"[b]{config_path}[/b] you can adjust it's settings there and"
                " restart the compose."
            )
        )

        if args.systemd_service:

            service_file.write_text(
                SYSTEMD_TMPL.format(
                    project_name=DF.project_name, engine=args.container_engine
                )
            )
        RichConsole.print(
            (
                "\n\nA systemd service file was created. Set the path"
                " to the compose file on the server and place it "
                f"into [b]/etc/systemd/system/{service_file.name}[/]"
                "\n"
                "then use:\n\n"
                "  [b]sudo systemctl daemon-reload\n"
                f"  sudo systemctl enable --now {service_file.name}[/b]\n"
            )
        )


def compose_parser(
    epilog: str = "", parser: Optional[argparse.ArgumentParser] = None
) -> None:
    """Construct command line argument parser."""
    parser = parser or argparse.ArgumentParser(
        prog="deploy-freva-compose",
        description="Create and inspect freva configuration.",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
        epilog=epilog,
    )
    parser.add_argument(
        "-v", "--verbose", action="count", help="Verbosity level", default=0
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    parser.add_argument(
        "-c",
        "--config-file",
        type=Path,
        help="Path to ansible inventory file.",
        default=config_dir / "config" / "inventory.toml",
    )
    parser.add_argument(
        "--host",
        type=str,
        help="Host name where the compose service should be running.",
    )
    parser.add_argument(
        "-u",
        "--user",
        type=str,
        help="User name that should run the services insight the container.",
        default="root",
    )
    parser.add_argument(
        "--no-plugins",
        action="store_true",
        help="Do not setup core library to use plugins.",
    )
    parser.add_argument(
        "-e",
        "--container-engine",
        help="Create a compose file for docker or podman.",
        default="docker",
        choices=["docker", "podman"],
        type=str,
    )
    parser.add_argument(
        "-s",
        "--systemd-service",
        help="Create a systemd-service file.",
        action="store_true",
    )
    parser.set_defaults(cli=create_compose)
